"""
PC-BASIC - compat.python2
Python 2 backports for Python 3 functionality

Contains functions from Python 3.3 source code, which is
copyright (c) 2001-2016 Python Software Foundation
and released under a GPL-compatible licence https://docs.python.org/3.3/license.html

Contains lines of code from package six, which is
Copyright (c) 2010-2018 Benjamin Peterson
and released under an MIT licence https://opensource.org/licenses/MIT
"""

# pylint: disable=import-error, no-member, undefined-variable

import shutil
import codecs
import contextlib
import itertools
import tempfile
import socket
import errno
import sys
import os

import ConfigParser as configparser
import Queue as queue
import copy_reg as copyreg

_FS_ENCODING = sys.getfilesystemencoding()


# __path__ hack for __init__ to ensure os.chdir does not break intra-package imports
# which they do because the package __path__ is given relative to cwd
# at least if run with python2 -m package
from .. import __path__
__path__[:] = [os.path.abspath(_e) for _e in __path__]


# deal with broken pipes

BrokenPipeError = IOError

def is_broken_pipe(e):
    return isinstance(e, IOError) and e.errno == errno.EPIPE


# strings

int2byte = chr

def iterchar(s):
    """Iterate over bytes, returning char."""
    return s

def iterbytes(s):
    """Iterate over bytes/bytearray/memoryview, returning int."""
    if isinstance(s, (bytes, memoryview)):
        return (ord(_c) for _c in s)
    return s

def add_str(cls):
    """Decorator to implement the correct str() function."""
    try:
        cls.__str__ = cls.__bytes__
    except AttributeError:
        pass
    return cls


# unicode system interfaces

getcwdu = os.getcwdu

# following python 3.5 this uses sys.getfilesystemencoding()
def getenvu(key, default=None):
    assert isinstance(key, unicode), type(key)
    try:
        return os.environ[key.encode(_FS_ENCODING)].decode(_FS_ENCODING)
    except KeyError:
        return default

def setenvu(key, value):
    assert isinstance(key, unicode), type(key)
    assert isinstance(value, unicode), type(value)
    os.environ[key.encode(_FS_ENCODING)] = value.encode(_FS_ENCODING)

def iterenvu():
    return (_key.decode(_FS_ENCODING) for _key in os.environ)


# iterators

xrange = xrange
zip = itertools.izip

def iteritems(d, **kw):
    return d.iteritems(**kw)

def itervalues(d, **kw):
    return d.itervalues(**kw)

def iterkeys(d, **kw):
    return d.iterkeys(**kw)

# utilities

# from Python3.3 shutil module source
def which(cmd, mode=os.F_OK | os.X_OK, path=None):
    """Given a command, mode, and a PATH string, return the path which
    conforms to the given mode on the PATH, or None if there is no such
    file.

    `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
    of os.environ.get("PATH"), or can be overridden with a custom search
    path.

    """
    # Check that a given file can be accessed with the correct mode.
    # Additionally check that `file` is not a directory, as on Windows
    # directories pass the os.access check.
    def _access_check(fn, mode):
        return (os.path.exists(fn) and os.access(fn, mode)
                and not os.path.isdir(fn))

    # If we're given a path with a directory part, look it up directly rather
    # than referring to PATH directories. This includes checking relative to the
    # current directory, e.g. ./script
    if os.path.dirname(cmd):
        if _access_check(cmd, mode):
            return cmd
        return None

    if path is None:
        path = os.environ.get("PATH", os.defpath)
    if not path:
        return None
    path = path.split(os.pathsep)

    if sys.platform == "win32":
        # The current directory takes precedence on Windows.
        if not os.curdir in path:
            path.insert(0, os.curdir)

        # PATHEXT is necessary to check on Windows.
        pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
        # See if the given file matches any of the expected path extensions.
        # This will allow us to short circuit when given "python.exe".
        # If it does match, only test that one, otherwise we have to try
        # others.
        if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
            files = [cmd]
        else:
            files = [cmd + ext for ext in pathext]
    else:
        # On other platforms you don't have things like PATHEXT to tell you
        # what file suffixes are executable, so just pass on cmd as-is.
        files = [cmd]

    seen = set()
    for dir in path:
        normdir = os.path.normcase(dir)
        if not normdir in seen:
            seen.add(normdir)
            for thefile in files:
                name = os.path.join(dir, thefile)
                if _access_check(name, mode):
                    return name
    return None


class SimpleNamespace(object):
    """Namespace with attribute access, like in Python 3
    https://docs.python.org/dev/library/types.html#types.SimpleNamespace
    """

    def __init__(self, **kwargs):
        """Initialise the namespace with keyword arguments."""
        self.__dict__.update(kwargs)

    def __repr__(self):
        """Bytes representation (for Python 2)."""
        keys = sorted(self.__dict__)
        items = (b'{}={!r}'.format(k, self.__dict__[k]) for k in keys)
        return b'{}({})'.format(type(self).__name__, b', '.join(items))

    def __eq__(self, other):
        """Namespaces are equal if their entries are equal."""
        return self.__dict__ == other.__dict__


class TemporaryDirectory():
    """Temporary directory context guard like in Python 3 tempfile."""

    def __init__(self, prefix=u''):
        """Initialise context guard."""
        self._prefix = prefix
        self._temp_dir = None

    def __enter__(self):
        """Create temp directory."""
        self._temp_dir = tempfile.mkdtemp(prefix=self._prefix)
        return self._temp_dir

    def __exit__(self, dummy_1, dummy_2, dummy_3):
        """Clean up temp directory."""
        if self._temp_dir:
            try:
                shutil.rmtree(self._temp_dir)
            except EnvironmentError as e:
                pass



"""
PC-BASIC - display.display
Display and video mode operations

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct
import logging

from ...compat import iteritems
from ..base import signals
from ..base import error
from .. import values
from . import graphics
from . import modes
from . import font

from .buffers import VideoBuffer
from .textscreen import TextScreen
from .colours import MONO_TINT
from .cursor import Cursor


class Display(object):
    """Display and video mode operations."""

    def __init__(
            self, queues, values, input_methods, memory,
            initial_width, video_mem_size, adapter, monitor,
            codepage, fonts
        ):
        """Initialise the display."""
        self._queues = queues
        self._values = values
        self._memory = memory
        # low level settings
        if adapter == 'ega':
            if monitor in MONO_TINT:
                adapter = 'ega_mono'
            elif video_mem_size < 131072:
                # less than 128k means 64k, as EGA has memory in 64k blocks
                adapter = 'ega_64k'
        self._adapter = adapter
        self._monitor = monitor
        self._video_mem_size = video_mem_size
        # prepare video modes
        self.mode = modes.get_mode(
            0, initial_width, self._adapter, self._monitor, self._video_mem_size
        )
        # video mode settings
        self.colorswitch, self.apagenum, self.vpagenum = 1, 0, 0
        # current attribute
        self.attr = self.mode.attr
        # border attribute
        self._border_attr = 0
        # prepare fonts
        self._fonts = {}
        if fonts:
            self._fonts = {
                height: font.Font(height, font_dict, codepage)
                for height, font_dict in iteritems(fonts) if font_dict
            }
        # we must have an 8-pixel font; use the default CP437 font if none provided
        # but note that we interpret the characters through the codepage provided
        if 8 not in self._fonts:
            self._fonts[8] = font.Font(8, None, codepage)
        # copy as 8-pixel hardware BIOS font (for CGA textmodes)
        # as opposed to the loadable 8-pixel memory font used in graphics modes
        self._bios_font_8 = self._fonts[8].copy()
        # text screen
        self._codepage = codepage
        self.cursor = Cursor(queues, self.mode)
        self.text_screen = TextScreen(self._values, self.mode, self.cursor, self._adapter)
        # page buffers, set by _set_mode
        self.pages = None
        # screen aspect ratio: used to determine pixel aspect ratio, which is used by CIRCLE
        # all adapters including PCjr target 4x3, except Tandy
        if self._adapter == 'tandy':
            aspect = (3072, 2000)
        else:
            aspect = (4, 3)
        # Tandy pixel aspect ratio is different from normal
        # suggesting screen aspect ratio is not 4/3.
        # Tandy pixel aspect ratios, experimentally found with CIRCLE (on DOSBox?):
        # screen 2, 6:     48/100   normal if aspect = 3072, 2000
        # screen 1, 4, 5:  96/100   normal if aspect = 3072, 2000
        # pcjr: screen 1: .833, s2: .833/2, s3: .833*2 s4:.833 s5:.833 s6: .833/2
        # (checked with CIRCLE on MAME)
        # .833 == 5:6 corresponding to screen aspect ratio of 4:3
        # --> old value SCREEN 3 pixel aspect 1968/1000 not quite (but almost) consistent with this
        #     and I don't think it was really checked on Tandy -- dosbox won't run SCREEN 3
        #
        # colour palette
        self.colourmap = self.mode.colourmap(
            self._queues, self._adapter, self._monitor, self.colorswitch
        )
        # graphics operations
        self.graphics = graphics.Graphics(
            input_methods, self._values, self._memory, aspect, self.colourmap
        )
        # initialise a fresh textmode screen
        self._set_mode(self.mode, 1, 0, 0, erase=True)

    @property
    def apage(self):
        """Active-page video buffers."""
        return self.pages[self.apagenum]

    @property
    def vpage(self):
        """Visible-page video buffers."""
        return self.pages[self.vpagenum]

    ###########################################################################
    # video modes

    def screen(
            self, new_mode_nr, new_colorswitch, new_apagenum, new_vpagenum,
            erase=1, new_width=None, force_reset=False
        ):
        """Change the video mode, colourburst, visible or active page."""
        # reset palette happens even if the SCREEN call fails
        self.colourmap.reset()
        # find the new mode we're trying to get into
        if new_mode_nr is None:
            # keep current mode if graphics but maybe change width if text
            if self.mode.is_text_mode and new_width is not None:
                new_mode = modes.get_mode(
                    0, new_width, self._adapter, self._monitor, self._video_mem_size
                )
            else:
                new_mode = self.mode
        else:
            if new_mode_nr == 0 and new_width is None:
                # if we switch out of a 20-col mode (Tandy screen 3), switch to 40-col.
                # otherwise, width persists on change to screen 0
                new_width = 40 if (self.mode.width == 20) else self.mode.width
            # retrieve the specs for the new video mode
            new_mode = modes.get_mode(
                new_mode_nr, new_width, self._adapter, self._monitor, self._video_mem_size
            )
        # set colorswitch
        new_colorswitch = bool(new_colorswitch)
        if new_colorswitch is None:
            new_colorswitch = True
            if self._adapter == 'pcjr':
                new_colorswitch = False
            elif self._adapter == 'tandy':
                new_colorswitch = new_mode.is_text_mode
        # vpage and apage nums are persistent on mode switch with SCREEN
        # on pcjr only, reset page to zero if current page number would be too high.
        # in other adapters, that's going to raise an IFC later on.
        if new_apagenum is None:
            new_apagenum = self.apagenum
            if (self._adapter == 'pcjr' and new_apagenum >= new_mode.num_pages):
                new_apagenum = 0
        if new_vpagenum is None:
            new_vpagenum = new_apagenum
            if (self._adapter == 'pcjr' and new_vpagenum >= new_mode.num_pages):
                new_vpagenum = 0
        # if the new mode has fewer pages than current vpage/apage
        # illegal fn call before anything happens.
        # signal the signals to change the screen resolution
        if (new_apagenum >= new_mode.num_pages or new_vpagenum >= new_mode.num_pages):
            raise error.BASICError(error.IFC)
        # if mode or colorswitch changed, do a full reset
        # otherwise only change pages
        if force_reset or new_mode != self.mode or new_colorswitch != self.colorswitch:
            self._set_mode(new_mode, new_colorswitch, new_apagenum, new_vpagenum, erase)
        else:
            self.set_page(new_vpagenum, new_apagenum)

    def _set_mode(self, new_mode, new_colorswitch, new_apagenum, new_vpagenum, erase):
        """Change the video mode, colourburst, visible or active page."""
        # preserve memory if erase==0; don't distingush erase==1 and erase==2
        if not erase:
            saved_addr, saved_buffer = self.mode.memorymap.get_all_memory(self)
        # switching to another text mode (width-only change or no change to text mode)
        text_to_text = self.mode.is_text_mode and new_mode.is_text_mode
        # mode is changing
        mode_changes = new_mode != self.mode
        colorswitch_changes = new_colorswitch != self.colorswitch
        page_changes = (new_vpagenum, new_apagenum) != (self.vpagenum, self.apagenum)
        # set the screen mode parameters
        self.mode = new_mode
        # set the colorswitch
        self.colorswitch = new_colorswitch
        # get a font for the new mode
        try:
            # in CGA (8- or 9-pixel) text modes, use the 8-pixel BIOS font, not the POKE-able one
            # 9th pixel is added by rendering code if needed
            if new_mode.is_text_mode and new_mode.font_height in (8, 9):
                font = self._bios_font_8
            else:
                font = self._fonts[new_mode.font_height]
            # initialise for this mode's font width and height
            # which is important for wdth or height 9 pixels
            font.init_mode(new_mode.font_width, new_mode.font_height)
        except KeyError:
            logging.warning(
                'No %d-pixel font available. Using 8-pixel font instead.',
                new_mode.font_height
            )
            font = self._bios_font_8.init_mode(new_mode.font_width, new_mode.font_height)
        # attribute and border persist on width-only change
        # otherwise start with black border and default attr
        if (not text_to_text or page_changes or colorswitch_changes):
            self.attr = new_mode.attr
        # initialise the palette
        self.colourmap = new_mode.colourmap(
            self._queues, self._adapter, self._monitor, self.colorswitch
        )
        # initialise pixel and character buffers
        self.pages = [
            VideoBuffer(
                self._queues,
                self.mode.pixel_height, self.mode.pixel_width,
                self.mode.height, self.mode.width,
                self.colourmap, self.attr, font, self._codepage,
                do_fullwidth=(self.mode.is_text_mode and self.mode.font_height >= 14),
            )
            for _pagenum in range(self.mode.num_pages)
        ]
        # submit the mode change to the interface
        self._queues.video.put(signals.Event(
            signals.VIDEO_SET_MODE, (
                new_mode.pixel_height, new_mode.pixel_width,
                new_mode.height, new_mode.width
            )
        ))
        # border persists on width-only change or no change; set to black otherwise
        if not text_to_text and mode_changes:
            self.set_border(0)
        # rebuild the cursor
        if not self.mode.is_text_mode and self.mode.cursor_attr:
            self.cursor.init_mode(self.mode, self.mode.cursor_attr, self.colourmap)
        else:
            self.cursor.init_mode(self.mode, self.attr, self.colourmap)
        # initialise text screen
        self.text_screen.init_mode(self.mode, self.pages, self.attr, new_vpagenum, new_apagenum)
        # restore emulated video memory in new mode
        if not erase:
            self.mode.memorymap.set_memory(self, saved_addr, saved_buffer)
        # center graphics cursor, reset window, etc.
        self.graphics.init_mode(self.mode, self.pages, self.colourmap)
        # set active page & visible page, counting from 0.
        self.set_page(new_vpagenum, new_apagenum)
        # set graphics attribute
        self.graphics.set_attr(self.attr)

    def set_width(self, to_width):
        """Set the number of columns of the screen, reset pages and change modes."""
        # if we're currently at that width, do nothing
        if to_width == self.mode.width:
            return
        new_mode = modes.to_width(self._adapter, self.mode, to_width)
        self.screen(new_mode, None, 0, 0, new_width=to_width)

    def set_height(self, to_height):
        """Try to change the number of rows."""
        # number != 25 is ignored on tandy, error elsewhere
        # otherwise nothing happens
        if self._adapter in ('pcjr', 'tandy'):
            error.range_check(0, 25, to_height)
        else:
            error.range_check(25, 25, to_height)

    def set_video_memory_size(self, new_size):
        """Change the amount of memory available to the video card."""
        new_size = int(new_size)
        if self._video_mem_size == new_size:
            return
        self._video_mem_size = new_size
        # reload max number of pages; do we fit? if not, drop to text
        self.mode.memorymap.set_video_mem_size(self._video_mem_size)
        # if not a no-op, rebuild pages and drop to text
        # note that we *must* rebuild the list of pages if we have more pages
        # and also we must do something if we have fewer pages and vpage or apage are currently on a higher page.
        # apparently pcjr always drops to text when this is not a no-op
        self.screen(0, 0, 0, 0, force_reset=True)

    def rebuild(self):
        """Completely resubmit the screen to the interface."""
        # set the screen mode
        self._queues.video.put(signals.Event(
            signals.VIDEO_SET_MODE, (
                self.mode.pixel_height, self.mode.pixel_width,
                self.mode.height, self.mode.width
            )
        ))
        # rebuild palette
        self.colourmap.submit()
        # set the border
        self._queues.video.put(signals.Event(signals.VIDEO_SET_BORDER_ATTR, (self._border_attr,)))
        # redraw the text screen and submit to interface
        for page in self.pages:
            page.resubmit()
        # rebuild cursor
        # this should come *after* submission of the contents
        # as video_cli waits for cursor position changes to show content
        self.cursor.rebuild()

    ###########################################################################
    # memory accessible properties

    def get_mode_info_byte(self):
        """Screen mode info byte in low memory address 1125."""
        # blink vs bright
        # this can in theory be set with OUT 984 (&h3d8), but not implemented
        blink_enabled = True
        # bit 0: only in text mode?
        # bit 2: should this be colorswitch or colorburst_is_enabled?
        return (
            (self.mode.is_text_mode and self.mode.width == 80) * 1 +
            (not self.mode.is_text_mode) * 2 +
            self.colorswitch * 4 + 8 +
            (self.mode.is_cga_hires) * 16 +
            blink_enabled * 32
        )

    def get_colour_info_byte(self):
        """Colour info byte in low memory address 1126."""
        if self.mode.is_text_mode:
            return self.get_border_attr()
            # not implemented: + 16 "if current color specified through
            # COLOR f,b with f in [0,15] and b > 7
        else:
            self.colourmap.get_colour_info_byte()

    @property
    def memory_font(self):
        """8-bit memory font (half in ROM, half in RAM and loadable)."""
        return self._fonts[8]

    @property
    def is_monochrome(self):
        """Check for monochrome monitor."""
        return self._monitor == 'mono'

    ###########################################################################

    def set_page(self, new_vpagenum, new_apagenum):
        """Set active page & visible page, counting from 0."""
        if new_vpagenum is None:
            new_vpagenum = self.vpagenum
        if new_apagenum is None:
            new_apagenum = self.apagenum
        if (new_vpagenum >= self.mode.num_pages or new_apagenum >= self.mode.num_pages):
            raise error.BASICError(error.IFC)
        try:
            self.pages[self.vpagenum].set_visible(False)
        except IndexError:
            # the page has been discarded, no need to set it invisible
            pass
        self.pages[new_vpagenum].set_visible(True)
        self.vpagenum = new_vpagenum
        self.apagenum = new_apagenum
        self.graphics.set_page(new_apagenum)
        self.text_screen.set_page(new_vpagenum, new_apagenum)

    def set_attr(self, attr):
        """Set the default attribute."""
        self.attr = attr
        self.graphics.set_attr(attr)
        self.text_screen.set_attr(attr)
        if not self.mode.is_text_mode and self.mode.cursor_attr is None:
            self.cursor.set_attr(attr)

    def set_border(self, attr):
        """Set the border attribute."""
        fore, _, _, _ = self.colourmap.split_attr(attr)
        self._border_attr = fore
        self._queues.video.put(signals.Event(signals.VIDEO_SET_BORDER_ATTR, (fore,)))

    def get_border_attr(self):
        """Get the border attribute, in range 0 <= attr < 16."""
        return self._border_attr


    ###########################################################################
    # callbacks

    # SCREEN statement
    # Colorswitch parameter
    #   SCREEN 0,0 - mono on composite, color on RGB
    #   SCREEN 0,1 - color (colorburst=True)
    #   SCREEN 1,0 - color (colorburst=True)
    #   SCREEN 1,1 - mono on composite, mode 5 on RGB
    # default colorswitch:
    #   SCREEN 0 = SCREEN 0,0 (pcjr)
    #   SCREEN 0 = SCREEN 0,1 (tandy, cga, ega, vga, ..)
    #   SCREEN 1 = SCREEN 1,0 (pcjr, tandy)
    #   SCREEN 1 = SCREEN 1,1 (cga, ega, vga, ...)
    # colorswitch is NOT preserved between screens when unspecified
    # colorswitch is NOT the same as colorburst (opposite on screen 1)
    #
    # Erase parameter:
    #   tells basic how much video memory to erase
    #   0: do not erase video memory
    #   1: (default) erase old and new page if screen or width changes
    #   2: erase all video memory if screen or width changes
    #   -> we're not distinguishing between 1 and 2 here

    def screen_(self, args):
        """SCREEN: change the video mode, colourburst, visible or active page."""
        # in GW, screen 0,0,0,0,0,0 raises error after changing the palette
        # this raises error before
        mode, colorswitch, apagenum, vpagenum = (
            None if arg is None else values.to_int(arg)
            for _, arg in zip(range(4), args)
        )
        # if any parameter not in [0,255], error 5 without doing anything
        # if the parameters are outside narrow ranges
        # (e.g. not implemented screen mode, pagenum beyond max)
        # then the error is only raised after changing the palette.
        error.range_check(0, 255, mode, colorswitch, apagenum, vpagenum)
        if self._adapter == 'tandy':
            error.range_check(0, 1, colorswitch)
        erase = next(args)
        if erase is not None:
            erase = values.to_int(erase)
            error.range_check(0, 2, erase)
        list(args)
        if erase is not None:
            # erase can only be set on pcjr/tandy 5-argument syntax
            if self._adapter not in ('pcjr', 'tandy'):
                raise error.BASICError(error.IFC)
        else:
            erase = 1
        self.screen(mode, colorswitch, apagenum, vpagenum, erase)

    def pcopy_(self, args):
        """Copy source to destination page."""
        src = values.to_int(next(args))
        error.range_check(0, self.mode.num_pages-1, src)
        dst = values.to_int(next(args))
        list(args)
        error.range_check(0, self.mode.num_pages-1, dst)
        self.pages[dst].copy_from(self.pages[src])

    def color_(self, args):
        """COLOR: set colour attributes."""
        args = list(args)
        error.throw_if(len(args) > 3)
        args += [None] * (3 - len(args))
        arg0, arg1, arg2 = args
        arg0 = arg0 and values.to_int(arg0)
        arg1 = arg1 and values.to_int(arg1)
        arg2 = arg2 and values.to_int(arg2)
        attr, border = self.colourmap.color(self.attr, arg0, arg1, arg2)
        if attr is not None:
            self.set_attr(attr)
        if border is not None:
            self.set_border(border)

    def palette_(self, args):
        """PALETTE: assign colour to attribute."""
        attrib = next(args)
        if attrib is not None:
            attrib = values.to_int(attrib)
        colour = next(args)
        if colour is not None:
            colour = values.to_int(colour)
        list(args)
        # wait a tick to make colour cycling loops work
        self._queues.wait()
        if attrib is None and colour is None:
            self.colourmap.set_all(self.colourmap.default_palette)
        else:
            # can't set blinking colours separately
            error.range_check(0, self.colourmap.num_palette-1, attrib)
            # numbers 255 and up are in fact allowed, 255 -> -1, 256 -> 0, etc
            colour = -1 + (colour + 1) % 256
            error.range_check(-1, self.colourmap.num_colours-1, colour)
            if colour != -1:
                self.colourmap.set_entry(attrib, colour)

    def palette_using_(self, args):
        """PALETTE USING: set palette from array buffer."""
        array_name, start_indices = next(args)
        array_name = self._memory.complete_name(array_name)
        list(args)
        try:
            dimensions = self._memory.arrays.dimensions(array_name)
        except KeyError:
            raise error.BASICError(error.IFC)
        error.throw_if(array_name[-1:] != values.INT, error.TYPE_MISMATCH)
        lst = self._memory.arrays.view_full_buffer(array_name)
        start = self._memory.arrays.index(start_indices, dimensions)
        num_palette_entries = self.colourmap.num_palette
        error.throw_if(self._memory.arrays.flat_length(dimensions) - start < num_palette_entries)
        new_palette = []
        for i in range(num_palette_entries):
            offset = (start+i) * 2
            ## signed int, as -1 means don't set
            val, = struct.unpack('<h', lst[offset:offset+2])
            error.range_check(-1, self.colourmap.num_colours-1, val)
            new_palette.append(val if val > -1 else self.colourmap.get_entry(i))
        self.colourmap.set_all(new_palette)

    def cls_(self, args):
        """CLS: clear the screen."""
        val = next(args)
        if val is not None:
            val = values.to_int(val)
            # tandy gives illegal function call on CLS number
            error.throw_if(self._adapter == 'tandy')
            error.range_check(0, 2, val)
        list(args)
        # CLS is only executed if no errors have occurred
        if not self.mode.is_text_mode and (
                val == 1 or (val is None and self.graphics.graph_view.active)
            ):
            # CLS 1: in graphics mode, clear the graphics viewport
            _, back, _, _ = self.colourmap.split_attr(self.attr)
            self.graphics.graph_view[:, :] = back
            self.graphics.reset()
            if not self.graphics.graph_view.active:
                self.text_screen.redraw_bar()
                self.text_screen.set_pos(1, 1)
        elif val == 0 or (val is None and not self.text_screen.scroll_area.active):
            self.text_screen.clear()
            self.text_screen.redraw_bar()
            self.graphics.reset()
        elif val == 2:
            # CLS 2 does not reset the graphics pointer (checked with DOSBox)
            # if vscroll area is not active, this does not clear the bottom bar
            self.text_screen.clear_view()
        elif val is None:
            # however, CLS only with active scroll area *does* reset the view
            # clear_view will clear the whole screen if the view is not set
            self.text_screen.clear_view()
            self.graphics.reset()



"""
SDL2 library loader.
Locate and link the required SDL2 dlls for use by sdl2.py
"""

"""
This file contains code from Marcus von Appen's pysdl2.
The original package is at https://github.com/py-sdl/py-sdl2

pysdl2 licence
==============
This software is distributed under the Public Domain. Since it is
not enough anymore to tell people: 'hey, just do with it whatever
you like to do', you can consider this software being distributed
under the CC0 Public Domain Dedication
(http://creativecommons.org/publicdomain/zero/1.0/legalcode.txt).

In cases, where the law prohibits the recognition of Public Domain
software, this software can be licensed under the zlib license as
stated below:

Copyright (C) 2012-2018 Marcus von Appen <marcus@sysfault.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software
   in a product, an acknowledgement in the product documentation would be
   appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
"""

import sys
import os
import warnings
import logging
from platform import machine as cpu_arch
from ctypes import CDLL
from ctypes.util import find_library

try:
    # sdl2dll will set the SDL2_DLL_PATH environment variable
    # unlike pysdl2, we do not use this variable
    with warnings.catch_warnings():
        # ignore warnings about the source package
        # we only want to make noise if we actually want to use SDL2 but can't link the library
        warnings.simplefilter('ignore')
        import sdl2dll
except ImportError:
    sdl2dll = None


# from pysdl2's dll.py

def _finds_libs_at_path(libnames, path, patterns):
    """Find libraries matching a given name (e.g. SDL2) in a specific path.
    """
    # Adding the potential 'd' suffix that is present on the library
    # when built in debug configuration
    searchfor = libnames + [libname + 'd' for libname in libnames]
    results = []

    # First, find any libraries matching pattern exactly within given path
    for libname in searchfor:
        for subpath in str.split(path, os.pathsep):
            for pattern in patterns:
                dllfile = os.path.join(subpath, pattern.format(libname))
                if os.path.exists(dllfile):
                    results.append(dllfile)

    # Next, on Linux and similar, find any libraries with version suffixes matching
    # pattern (e.g. libSDL2.so.2) at path and add them in descending version order
    # (i.e. newest first)
    if sys.platform not in ("win32", "darwin"):
        versioned = []
        files = os.listdir(path)
        for f in files:
            for libname in searchfor:
                dllname = "lib{0}.so".format(libname)
                if dllname in f and not (dllname == f or f.startswith(".")):
                    versioned.append(os.path.join(path, f))
        versioned.sort(key = _so_version_num, reverse = True)
        results = results + versioned

    return results

def _so_version_num(libname):
    """Extracts the version number from an .so filename as a list of ints."""
    return list(map(int, libname.split('.so.')[1].split('.')))


def _findlib(libnames, paths):
    """Find libraries with a given name and return their paths in a list.
    If a path is specified, libraries found in that directory will take precedence,
    with libraries found in system search paths following.
    """

    platform = sys.platform
    if platform == "win32":
        patterns = ["{0}.dll"]
    elif platform == "darwin":
        patterns = ["lib{0}.dylib", "{0}.framework/{0}", "{0}.framework/Versions/A/{0}"]
    else:
        patterns = ["lib{0}.so"]

    # Adding the potential 'd' suffix that is present on the library
    # when built in debug configuration
    searchfor = libnames + [libname + 'd' for libname in libnames]

    # First, find any matching libraries at the given path (if specified)
    results = []
    for path in paths:
        if os.path.exists(path) and path.lower() != "system":
            results = _finds_libs_at_path(libnames, path, patterns)

    # Next, search for library in system library search paths
    for libname in searchfor:
        dllfile = find_library(libname)
        if dllfile:
            # For Python 3.8+ on Windows, need to specify relative or full path
            if os.name == "nt" and not ("/" in dllfile or "\\" in dllfile):
                dllfile = "./" + dllfile
            results.append(dllfile)

    # On ARM64 Macs, search the non-standard brew library path as a fallback
    arm_brewpath = "/opt/Homebrew/lib"
    is_apple_silicon = platform == "darwin" and cpu_arch() == "arm64"
    if is_apple_silicon and os.path.exists(arm_brewpath):
        results += _finds_libs_at_path(libnames, arm_brewpath, patterns)

    return results


class DLLWarning(Warning):
    pass


class DLL(object):
    """
    Function wrapper around the different DLL functions. Do not use or
    instantiate this one directly from your user code.
    """
    def __init__(self, libinfo, libnames, paths):
        self._dll = None
        foundlibs = _findlib(list(libnames), paths)
        if not foundlibs:
            if paths:
                pathinfo = '(looked in paths: {})'.format(paths)
            raise RuntimeError("could not find any library for {0} {1}".format(libinfo, pathinfo))
        for libfile in foundlibs:
            try:
                self._dll = CDLL(libfile)
                self._libfile = libfile
                self._libpath = os.path.dirname(libfile)
                break
            except Exception as exc:
                # Could not load the DLL, move to the next, but inform the user
                # about something weird going on - this may become noisy, but
                # is better than confusing the users with the RuntimeError below
                warnings.warn(repr(exc), DLLWarning)
        if self._dll is None:
            raise RuntimeError(
                "found %s, but it's not usable for the library %s" %  (foundlibs, libinfo)
            )
        # add library path to the PATH environment on Windows
        if self._libpath and sys.platform == 'win32':
            os.environ["PATH"] = "%s;%s" % (self._libpath, os.environ["PATH"])

    def bind_function(self, funcname, args=None, returns=None, optfunc=None):
        """
        Binds the passed argument and return value types to the specified
        function.
        """
        func = getattr(self._dll, funcname, None)
        warnings.warn\
            ("function '%s' not found in %r, using replacement" %
             (funcname, self._dll), ImportWarning)
        if not func:
            if optfunc:
                warnings.warn\
                    ("function '%s' not found in %r, using replacement" %
                     (funcname, self._dll), ImportWarning)
                func = _nonexistent(funcname, optfunc)
            else:
                raise ValueError("could not find function '%s' in %r" %
                                 (funcname, self._dll))
        func.argtypes = args
        func.restype = returns
        return func

    @property
    def libfile(self):
        """Gets the filename of the loaded library."""
        return self._libfile

def _nonexistent(funcname, func):
    """A simple wrapper to mark functions and methods as nonexistent."""
    def wrapper(*fargs, **kw):
        warnings.warn("%s does not exist" % funcname,
                      category=RuntimeWarning, stacklevel=2)
        return func(*fargs, **kw)
    wrapper.__name__ = func.__name__
    return wrapper


###################################################################################################

# DLL object linking to main SDL2 library
sdl2_lib = None
gfx_lib = None

# possible names of sdl and sdl_gfx library
SDL_NAMES = ('SDL2', 'SDL2-2.0')
GFX_NAMES = ('SDL2_gfx', 'SDL2_gfx-1.0')


def load_dlls(*library_paths):
    """
    Attempt to link to the required SDL2 dlls.
    This function must be called before importing the sdl2 module.
    """
    global sdl2_lib, gfx_lib

    # get modules from the override path first, then from sdl2dll package, if installed
    if sdl2dll:
        library_paths += (sdl2dll.get_dllpath(),)
    try:
        sdl2_lib = DLL('SDL2', SDL_NAMES, library_paths)
    except RuntimeError as exc:
        logging.warning('Failed to load library sdl2: %s' % exc)
    try:
        gfx_lib = DLL('SDL2_gfx', GFX_NAMES, library_paths)
    except RuntimeError as exc:
        logging.warning('Failed to load library sdl2_gfx: %s' % exc)



"""
PC-BASIC - audio_pygame.py
Sound interface based on PyGame

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import sys
from collections import deque


if False:
    # keep the import for detection by packagers
    import pygame
    import pygame.mixer as mixer

from ..basic.base import signals
from .audio import AudioPlugin
from .base import audio_plugins, InitFailed
from . import synthesiser


# one wavelength at 37 Hz is 1192 samples at 44100 Hz
CHUNK_LENGTH = 1192 * 4
# quit sound server after quiet period of QUIET_QUIT ticks
# to avoid high-ish cpu load from the sound server.
QUIET_QUIT = 10000
# buffer size in sample frames
BUFSIZE = 1024 #4096


##############################################################################
# plugin

@audio_plugins.register('pygame')
class AudioPygame(AudioPlugin):
    """Pygame-based audio plugin."""

    def __init__(self, audio_queue, **kwargs):
        """Initialise sound system."""
        global pygame, mixer
        try:
            import pygame
        except ImportError:
            raise InitFailed('Module `pygame` not found')
        try:
            from pygame import mixer
        except ImportError:
            raise InitFailed('Module `mixer` not found')
        # this must be called before pygame.init() in the video plugin
        # if sample_bits != 16 or -16 I get no sound. seems to ave no effect though
        mixer.pre_init(
            synthesiser.SAMPLE_RATE, -16, channels=1, buffer=BUFSIZE
        )
        # synthesisers
        self.signal_sources = synthesiser.get_signal_sources()
        # sound generators for each voice
        self.generators = [deque(), deque(), deque(), deque()]
        # do not quit mixer if true
        self._persist = False
        # keep track of quiet time to shut down mixer after a while
        self.quiet_ticks = 0
        AudioPlugin.__init__(self, audio_queue)

    def __enter__(self):
        """Perform any necessary initialisations."""
        # initialise mixer as silent
        # this is necessary to be able to set channels to mono
        mixer.quit()
        return AudioPlugin.__enter__(self)

    def persist(self, do_persist):
        """Allow or disallow mixer to quit."""
        self._persist = do_persist

    def tone(self, voice, frequency, duration, loop, volume):
        """Enqueue a tone."""
        self.generators[voice].append(synthesiser.SoundGenerator(
            self.signal_sources[voice], synthesiser.FEEDBACK_TONE,
            frequency, duration, loop, volume
        ))

    def noise(self, source, frequency, duration, loop, volume):
        """Enqueue a noise."""
        feedback = synthesiser.FEEDBACK_NOISE if source else synthesiser.FEEDBACK_PERIODIC
        self.generators[3].append(synthesiser.SoundGenerator(
            self.signal_sources[3], feedback,
            frequency, duration, loop, volume
        ))

    def hush(self):
        """Stop sound."""
        for voice in range(4):
            self._stop_channel(voice)
            self._next_tone[voice] = None
            while self.generators[voice]:
                self.generators[voice].popleft()

    def _work(self):
        """Replenish sample buffer."""
        if (
                sum(len(q) for q in self.generators) == 0 and
                self._next_tone == [None, None, None, None]
            ):
            # check if mixer can be quit
            self._check_quit()
            return
        self._check_init_mixer()
        for voice in range(4):
            if mixer.Channel(voice).get_queue() is not None:
                # nothing to do
                continue
            while True:
                if self._next_tone[voice] is None or self._next_tone[voice].loop:
                    try:
                        self._next_tone[voice] = self.generators[voice].popleft()
                    except IndexError:
                        if self._next_tone[voice] is None:
                            current_chunk = None
                            break
                current_chunk = self._next_tone[voice].build_chunk(CHUNK_LENGTH)
                if current_chunk is not None:
                    break
                self._next_tone[voice] = None
            if current_chunk is not None:
                # enqueue chunk in mixer
                snd = pygame.sndarray.make_sound(current_chunk)
                mixer.Channel(voice).queue(snd)

    def _check_quit(self):
        """Quit the mixer if not running a program and sound quiet for a while."""
        if self._next_tone != [None, None, None, None]:
            self.quiet_ticks = 0
        else:
            self.quiet_ticks += 1
            if not self._persist and self.quiet_ticks > QUIET_QUIT:
                # mixer is quiet and we're not running a program.
                # quit to reduce pulseaudio cpu load
                # this takes quite a while and leads to missed frames...
                if mixer.get_init() is not None:
                    mixer.quit()
                self.quiet_ticks = 0

    def _stop_channel(self, channel):
        """Stop sound on a channel."""
        if mixer.get_init():
            mixer.Channel(channel).stop()
            # play short silence to avoid blocking the channel
            # otherwise it won't play on queue()
            silence = pygame.sndarray.make_sound(bytearray(1))
            mixer.Channel(channel).play(silence)

    def _check_init_mixer(self):
        """Initialise the mixer if necessary."""
        if mixer.get_init() is None:
            mixer.init()



"""
PC-BASIC - error.py
Error constants and exceptions

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

# error constants
NEXT_WITHOUT_FOR = 1
SYNTAX_ERROR = 2
RETURN_WITHOUT_GOSUB = 3
OUT_OF_DATA = 4
ILLEGAL_FUNCTION_CALL = 5
OVERFLOW = 6
OUT_OF_MEMORY = 7
UNDEFINED_LINE_NUMBER = 8
SUBSCRIPT_OUT_OF_RANGE = 9
DUPLICATE_DEFINITION = 10
DIVISION_BY_ZERO = 11
ILLEGAL_DIRECT = 12
TYPE_MISMATCH = 13
OUT_OF_STRING_SPACE = 14
STRING_TOO_LONG = 15
STRING_FORMULA_TOO_COMPLEX = 16
CANT_CONTINUE = 17
UNDEFINED_USER_FUNCTION = 18
NO_RESUME = 19
RESUME_WITHOUT_ERROR = 20
# 21
MISSING_OPERAND = 22
LINE_BUFFER_OVERFLOW = 23
DEVICE_TIMEOUT = 24
DEVICE_FAULT = 25
FOR_WITHOUT_NEXT = 26
OUT_OF_PAPER = 27
# 28
WHILE_WITHOUT_WEND = 29
WEND_WITHOUT_WHILE = 30
# 31--49
FIELD_OVERFLOW = 50
INTERNAL_ERROR = 51
BAD_FILE_NUMBER = 52
FILE_NOT_FOUND = 53
BAD_FILE_MODE = 54
FILE_ALREADY_OPEN = 55
# 56
DEVICE_IO_ERROR = 57
FILE_ALREADY_EXISTS = 58
# 59--60
DISK_FULL = 61
INPUT_PAST_END = 62
BAD_RECORD_NUMBER = 63
BAD_FILE_NAME = 64
# 65
DIRECT_STATEMENT_IN_FILE = 66
TOO_MANY_FILES = 67
DEVICE_UNAVAILABLE = 68
COMMUNICATION_BUFFER_OVERFLOW = 69
PERMISSION_DENIED = 70
DISK_NOT_READY = 71
DISK_MEDIA_ERROR = 72
ADVANCED_FEATURE = 73
RENAME_ACROSS_DISKS = 74
PATH_FILE_ACCESS_ERROR = 75
PATH_NOT_FOUND = 76
DEADLOCK = 77

# shorthand
STX = SYNTAX_ERROR
IFC = ILLEGAL_FUNCTION_CALL


class Interrupt(Exception):
    """Base type for exceptions."""

    message = b''

    def __repr__(self):
        """String representation of exception."""
        return self.message.decode('ascii', 'replace')

    __str__ = __repr__

    def get_message(self, line_number=None):
        """Error message."""
        if line_number is not None and 0 <= line_number < 65535:
            return b'%s in %i' % (self.message, line_number)
        else:
            return b'%s' % (self.message,)


class Exit(Interrupt):
    """Exit emulator."""
    message = b'Exit'


class Reset(Exit):
    """Reset emulator."""
    message = b'Reset'


class Break(Interrupt):
    """Program interrupt."""

    message = b'Break'

    def __init__(self, stop=False, pos=None):
        """Initialise break."""
        Interrupt.__init__(self)
        self.stop = stop
        self.pos = pos
        self.trapped_error_num = None
        self.trapped_error_pos = None


class BASICError(Interrupt):
    """Runtime error."""

    default_message = b'Unprintable error'
    messages = {
        1: b'NEXT without FOR',
        2: b'Syntax error',
        3: b'RETURN without GOSUB',
        4: b'Out of DATA',
        5: b'Illegal function call',
        6: b'Overflow',
        7: b'Out of memory',
        8: b'Undefined line number',
        9: b'Subscript out of range',
        10: b'Duplicate Definition',
        11: b'Division by zero',
        12: b'Illegal direct',
        13: b'Type mismatch',
        14: b'Out of string space',
        15: b'String too long',
        16: b'String formula too complex',
        17: b"Can't continue",
        18: b'Undefined user function',
        19: b'No RESUME',
        20: b'RESUME without error',
        # 21
        22: b'Missing operand',
        23: b'Line buffer overflow',
        24: b'Device Timeout',
        25: b'Device Fault',
        26: b'FOR without NEXT',
        27: b'Out of paper',
        # 28
        29: b'WHILE without WEND',
        30: b'WEND without WHILE',
        # 31--49
        50: b'FIELD overflow',
        51: b'Internal error',
        52: b'Bad file number',
        53: b'File not found',
        54: b'Bad file mode',
        55: b'File already open',
        # 56
        57: b'Device I/O error',
        58: b'File already exists',
        # 59--60
        61: b'Disk full',
        62: b'Input past end',
        63: b'Bad record number',
        64: b'Bad file name',
        # 65
        66: b'Direct statement in file',
        67: b'Too many files',
        68: b'Device Unavailable',
        69: b'Communication buffer overflow',
        70: b'Permission Denied',
        71: b'Disk not Ready',
        72: b'Disk media error',
        73: b'Advanced Feature',
        74: b'Rename across disks',
        75: b'Path/File access error',
        76: b'Path not found',
        77: b'Deadlock',
    }

    def __init__(self, value, pos=None):
        """Initialise error."""
        Interrupt.__init__(self)
        self.err = value
        self.pos = pos
        try:
            self.message = self.messages[self.err]
        except KeyError:
            self.message = self.default_message


def range_check(lower, upper, *allvars):
    """Check if all variables in list are within the given inclusive range."""
    for v in allvars:
        if v is not None and not (lower <= v <= upper):
            raise BASICError(IFC)

def throw_if(bool, err=IFC):
    """Raise IFC if condition is met."""
    if bool:
        raise BASICError(err)

def range_check_err(lower, upper, v, err=IFC):
    """Check if variable is within the given inclusive range."""
    if v is not None and not (lower <= v <= upper):
        raise BASICError(err)



"""
PC-BASIC - inputs.keyboard
Keyboard handling

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from collections import deque
from contextlib import contextmanager

from ...compat import iterchar, int2byte

from ..base import error
from ..base import scancode
from ..base import signals
from ..base.eascii import as_bytes as ea
from ..base.eascii import as_unicode as uea


# bit flags for modifier keys - these are exposed via peek(1047) in low memory
# sticky modifiers
TOGGLE = {
    scancode.INSERT: 0x80, scancode.CAPSLOCK: 0x40,
    scancode.NUMLOCK: 0x20, scancode.SCROLLOCK: 0x10
}

# nonsticky modifiers
MODIFIER = {
    scancode.ALT: 0x8, scancode.CTRL: 0x4,
    scancode.LSHIFT: 0x2, scancode.RSHIFT: 0x1
}

# default function key eascii codes for KEY autotext.
FUNCTION_KEY = {
    ea.F1: 0, ea.F2: 1, ea.F3: 2, ea.F4: 3,
    ea.F5: 4, ea.F6: 5, ea.F7: 6, ea.F8: 7,
    ea.F9: 8, ea.F10: 9, ea.F11: 10, ea.F12: 11
}

# numeric keypad
KEYPAD = {
    scancode.KP0: b'0', scancode.KP1: b'1', scancode.KP2: b'2', scancode.KP3: b'3',
    scancode.KP4: b'4', scancode.KP5: b'5', scancode.KP6: b'6', scancode.KP7: b'7',
    scancode.KP8: b'8', scancode.KP9: b'9'
}

# function-key macros
DEFAULT_MACROS = (
    b'LIST ', b'RUN\r', b'LOAD"', b'SAVE"', b'CONT\r', b',"LPT1:"\r',
    b'TRON\r', b'TROFF\r', b'KEY ', b'SCREEN 0,0,0\r', b'', b''
)

# short beep (0.1s at 800Hz) emitted if buffer is full
FULL_TONE = (0, 800, 0.01, False, 15)


###############################################################################
# keyboard ring buffer

class KeyboardBuffer(object):
    """Quirky emulated ring buffer for keystrokes."""

    def __init__(self, queues, ring_length, check_full):
        """Initialise to given length."""
        self._queues = queues
        # buffer holds tuples (eascii/codepage, scancode, modifier)
        self._buffer = [(b'\0\0', 0)] * ring_length
        self._ring_length = ring_length
        self._start = ring_length
        # check if ring is full
        self._check_full = check_full

    @contextmanager
    def ignore_limit(self):
        """Enable/diable buffer limit check."""
        save, self._check_full = self._check_full, False
        yield
        self._check_full = save

    def append(self, cp_c, scan):
        """Append a single keystroke with eascii/codepage, scancode, modifier."""
        # if check_full is off, we pretend the ring buffer is infinite
        # this is for inserting keystrokes and pasting text into the emulator
        if cp_c:
            if self._check_full and len(self._buffer) - self._start >= self._ring_length-1:
                # when buffer is full, GW-BASIC inserts a \r at the end but doesn't count it
                self._buffer[self._start-1] = (b'\r', scancode.RETURN)
                # emit a sound signal; keystroke is dropped
                self._queues.audio.put(signals.Event(signals.AUDIO_TONE, FULL_TONE))
            else:
                self._buffer.append((cp_c, scan))

    def getc(self):
        """Read a keystroke as eascii/codepage."""
        try:
            c = self._buffer[self._start][0]
        except IndexError:
            c = b''
        else:
            self._start += 1
        return c

    def peek(self):
        """Show top keystroke in keyboard buffer as eascii/codepage."""
        try:
            return self._buffer[self._start][0]
        except IndexError:
            return b''

    def _ring_index(self, index):
        """Get index for ring position."""
        diff = len(self._buffer) % self._ring_length - index
        offset = len(self._buffer) - diff
        if diff <= 0:
            offset -= self._ring_length
        return offset

    @property
    def length(self):
        """Return the number of keystrokes in the buffer."""
        return min(self._ring_length, len(self._buffer) - self._start)

    @property
    def empty(self):
        """True if no keystrokes in buffer."""
        return self._start >= len(self._buffer)

    @property
    def start(self):
        """Ring buffer starting index."""
        return self._start % self._ring_length

    @property
    def stop(self):
        """Ring buffer stopping index."""
        return (self._start + self.length) % self._ring_length

    def ring_read(self, index):
        """Read character at position i in ring as eascii/codepage."""
        return self._buffer[self._ring_index(index)]

    def ring_write(self, index, c, scan):
        """Write character at position i in ring as eascii/codepage."""
        self._buffer[self._ring_index(index)] = (c, scan)

    def ring_set_boundaries(self, newstart, newstop):
        """Set start and stop index."""
        length = (newstop - newstart) % self._ring_length
        # rotate buffer to account for new start and stop
        # these are between length - ring_length and length
        start_index = self._ring_index(newstart)
        stop_index = self._ring_index(newstop)
        start = self._ring_index(self._start)
        # drop any extended buffer beyond ring limits
        self._buffer = self._buffer[:self._start + self._ring_length]
        # cut to ring limits, we should be exactly the right size
        start -= len(self._buffer) - self._ring_length
        self._buffer = self._buffer[-self._ring_length:]
        # rotate so that the stop index is at the end
        shift = len(self._buffer[start+length:])
        self._buffer = self._buffer[start+length:] + self._buffer[:start+length]
        start += shift
        start = start % self._ring_length
        # insert zeros before buffer to get the correct modulo
        while start % self._ring_length != newstart:
            start += 1
            self._buffer = [(b'\0\0', 0)] + self._buffer
        self._start = start


###############################################################################
# keyboard operations

def _iter_keystrokes(ueascii_seq):
    """Iterate over e-ascii/unicode sequences."""
    out = u''
    for char in ueascii_seq:
        if out or char != u'\0':
            yield out + char
            out = u''
        elif char == u'\0':
            # eascii code is \0 plus one char
            out = char


class Keyboard(object):
    """Keyboard handling."""

    def __init__(self, queues, values, codepage, check_full):
        """Initilise keyboard state."""
        self._values = values
        # needed for wait() in wait_char()
        self._queues = queues
        # key queue (holds bytes); ring buffer of length 16
        self.buf = KeyboardBuffer(queues, 16, check_full)
        # INP(&H60) scancode
        self.last_scancode = 0
        # active status of caps, num, scroll, alt, ctrl, shift modifiers
        self.mod = 0
        # store for alt+keypad ascii insertion
        self.keypad_ascii = b''
        # ignore caps lock, let OS handle it
        # this is now switched off hard-coded, but logic remains for now
        self._ignore_caps = True
        # pre-inserted keystrings
        self._codepage = codepage
        # stream buffer
        self._stream_buffer = deque()
        # redirected input stream has closed
        self._input_closed = False
        # expansion buffer for keyboard macros
        # expansion vessel holds codepage chars
        self._expansion_vessel = []
        # f-key macros
        self._key_replace = list(DEFAULT_MACROS)

    # event handler

    def check_input(self, signal):
        """Handle keyboard input signals and clipboard paste."""
        if signal.event_type == signals.KEYB_DOWN:
            # params is e-ASCII/unicode character sequence, scancode, modifier
            self._key_down(*signal.params)
        elif signal.event_type == signals.KEYB_UP:
            self._key_up(*signal.params)
        elif signal.event_type == signals.STREAM_CHAR:
            # params is a unicode sequence
            self._stream_chars(*signal.params)
        elif signal.event_type == signals.STREAM_CLOSED:
            self._close_input()
        elif signal.event_type == signals.CLIP_PASTE:
            self._stream_chars(*signal.params)
        else:
            return False
        return True

    def _key_down(self, c, scan, mods):
        """Insert a key-down event by eascii/unicode, scancode and modifiers."""
        if scan is not None:
            self.last_scancode = scan
        # update ephemeral modifier status at every keypress
        # mods is a list of scancodes; OR together the known modifiers
        self.mod &= ~(
            MODIFIER[scancode.CTRL] | MODIFIER[scancode.ALT] |
            MODIFIER[scancode.LSHIFT] | MODIFIER[scancode.RSHIFT]
        )
        if mods:
            for m in mods:
                self.mod |= MODIFIER.get(m, 0)
        # set toggle-key modifier status
        # these are triggered by keydown events
        try:
            self.mod ^= TOGGLE[scan]
        except KeyError:
            pass
        # alt+keypad ascii replacement
        if mods and (scancode.ALT in mods):
            try:
                self.keypad_ascii += KEYPAD[scan]
                return
            except KeyError:
                pass
        if (
                self.mod & TOGGLE[scancode.CAPSLOCK]
                and not self._ignore_caps and len(c) == 1
            ):
            c = c.swapcase()
        self.buf.append(self._codepage.unicode_to_bytes(c), scan)

    def _key_up(self, scan):
        """Insert a key-up event."""
        if scan is not None:
            self.last_scancode = 0x80 + scan
        try:
            # switch off ephemeral modifiers
            self.mod &= ~MODIFIER[scan]
        except KeyError:
           pass
        # ALT+keycode
        if scan == scancode.ALT and self.keypad_ascii:
            char = int2byte(int(self.keypad_ascii)%256)
            if char == b'\0':
                char = b'\0\0'
            self.buf.append(char, None)
            self.keypad_ascii = b''

    def _stream_chars(self, us):
        """Insert eascii/unicode string into stream buffer."""
        for ea_char in _iter_keystrokes(us):
            ea_cp = self._codepage.unicode_to_bytes(ea_char)
            # don't append empty strings arising from unknown unicode
            # as this will be interpreted as end of stream, terminating the interpreter
            if ea_cp:
                self._stream_buffer.append(ea_cp)

    def inject_keystrokes(self, keystring):
        """Insert eascii/unicode string into keyboard buffer."""
        with self.buf.ignore_limit():
            for ea_char in _iter_keystrokes(keystring):
                self.buf.append(self._codepage.unicode_to_bytes(ea_char), None)

    def _close_input(self):
        """Signal that input stream has closed."""
        self._input_closed = True

    # macros

    def set_macro(self, num, macro):
        """Set macro for given function key."""
        # NUL terminates macro string, rest is ignored
        # macro starting with NUL is empty macro
        self._key_replace[num-1] = macro.split(b'\0', 1)[0]

    def get_macro(self, num):
        """Get macro for given function key."""
        return self._key_replace[num]

    # character retrieval

    def wait_char(self, keyboard_only=False):
        """Block until character appears in keyboard queue or stream."""
        # if input stream has closed, don't wait but return empty
        # which will tell the Editor to close
        # except if we're waiting for KYBD: input
        while (
                (not self._expansion_vessel) and (self.buf.empty) and (
                    keyboard_only or (not self._input_closed and not self._stream_buffer)
                )
            ):
            self._queues.wait()

    def _read_kybd_byte(self, expand=True):
        """Read one byte from keyboard buffer, expanding macros if required."""
        try:
            return self._expansion_vessel.pop(0)
        except IndexError:
            pass
        c = self.buf.getc()
        if not expand or c not in FUNCTION_KEY:
            return c
        # function key macro expansion
        self._expansion_vessel = list(iterchar(self._key_replace[FUNCTION_KEY[c]]))
        try:
            return self._expansion_vessel.pop(0)
        except IndexError:
            # function macro has been redefined as empty: return scancode
            # e.g. KEY 1, "" enables catching F1 with INKEY$
            return c

    def inkey_(self, args):
        """INKEY$: read one byte from keyboard or stream; nonblocking."""
        list(args)
        inkey = self.read_byte()
        return self._values.new_string().from_str(inkey)

    def read_byte(self):
        """Read one byte from keyboard or stream; nonblocking."""
        # wait a tick to reduce load in loops
        self._queues.wait()
        inkey = self._read_kybd_byte()
        if not inkey and self._stream_buffer:
            inkey = self._stream_buffer.popleft()
        return inkey

    def read_bytes_block(self, n):
        """Read bytes from keyboard or stream; blocking."""
        word = []
        for _ in range(n):
            self.wait_char(keyboard_only=False)
            word.append(self.read_byte())
        return b''.join(word)

    def peek_byte_kybd_file(self):
        """Peek from keyboard only; for KYBD: files; blocking."""
        self.wait_char(keyboard_only=True)
        return self.buf.peek()

    def read_bytes_kybd_file(self, num):
        """Read num bytes from keyboard only; for KYBD: files; blocking."""
        word = []
        for _ in range(num):
            self.wait_char(keyboard_only=True)
            word.append(self._read_kybd_byte(expand=False))
        return word

    def get_fullchar(self, expand=True):
        """Read one (sbcs or dbcs) full character; nonblocking."""
        c = self._read_kybd_byte(expand)
        # insert dbcs chars from keyboard buffer two bytes at a time
        if (c in self._codepage.lead and self.buf.peek() in self._codepage.trail):
            c += self._read_kybd_byte(expand)
        if not c and self._stream_buffer:
            c = self._stream_buffer.popleft()
            if (c in self._codepage.lead and self._stream_buffer and
                        self._stream_buffer[0] in self._codepage.trail):
                c += self._stream_buffer.popleft()
        return c

    def get_fullchar_block(self, expand=True):
        """Read one (sbcs or dbcs) full character; blocking."""
        self.wait_char()
        return self.get_fullchar(expand)



"""
PC-BASIC - devices package
Devices, Files and I/O operations

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from .devicebase import TYPE_TO_MAGIC, InputTextFile
from .files import Files
from .disk import NameWrapper



"""
PC-BASIC - info
platform information

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import sys
import platform
import importlib

from .compat import which
from .basic import NAME, VERSION, LONG_VERSION, COPYRIGHT


def get_version_info():
    """Retrieve long version info."""
    return u'%s %s\n%s\n' % (NAME, LONG_VERSION, COPYRIGHT)

def get_platform_info():
    """Retrieve information about operating system and installed modules."""
    info = []
    info.append(u'\nPLATFORM')
    info.append(u'os: %s' % platform.platform())
    frozen = getattr(sys, 'frozen', '') or ''
    info.append(
        u'python: %s %s %s' % (
        sys.version.replace('\n', ''), ' '.join(platform.architecture()), frozen))
    info.append(u'\nMODULES')
    modules = ('pyaudio', 'serial', 'parallel')
    for module in modules:
        try:
            m = importlib.import_module(module)
        except Exception:
            info.append(u'%s: --' % module)
        else:
            for version_attr in ('__version__', 'version', 'VERSION'):
                try:
                    name = module.split('.')[-1]
                    version = getattr(m, version_attr)
                    if isinstance(version, bytes):
                        version = version.decode('ascii', 'ignore')
                    info.append(u'%s: %s' % (name, version))
                    break
                except AttributeError:
                    pass
            else:
                info.append(u'%s: available' % module)
    info.append(u'\nLIBRARIES')
    try:
        from .interface import video_sdl2
        video_sdl2._import_sdl2()

        info.append(u'sdl2: %s' % (video_sdl2.sdl2.sdl2_lib.libfile,))
        if video_sdl2:
            info.append(u'sdl2_gfx: %s' % (video_sdl2.sdl2.gfx_lib.libfile, ))
        else:
            info.append(u'sdl2_gfx: --')
    except ImportError as e:
        raise
        info.append(u'sdl2: --')
        sdl2 = None
    info.append(u'\nEXTERNAL TOOLS')
    tools = (u'notepad', u'lpr', u'paps', u'beep', u'pbcopy', u'pbpaste')
    for tool in tools:
        location = which(tool) or u'--'
        info.append(u'%s: %s' % (tool, location))
    info.append(u'')
    return u'\n'.join(info)



"""
PC-BASIC - iostreams.py
Input/output streams

(c) 2014--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging
import threading
import sys
import time
import io
from contextlib import contextmanager

from ..compat import WIN32, read_all_available, stdio, random_id, text_type
from .base import signals
from .codepage import CONTROL


# sleep period for input thread
# does not need to be very short as it reads multiple bytes in one cycle
TICK = 0.03


class IOStreams(object):
    """Manage input/output to files, printers and stdio."""

    def __init__(self, queues, codepage):
        """Initialise I/O streams."""
        self._queues = queues
        self._codepage = codepage
        # disable input streams at start
        self._active = False
        # flag for input daemon thread
        self._stop_threads = False
        # streams
        self._input_streams = []
        self._output_streams = []

    def __getstate__(self):
        """Pickle the streams."""
        return self.__dict__

    def __setstate__(self, pickle_dict):
        """Unpickle and resume the streams."""
        self.__dict__.update(pickle_dict)
        if self._input_streams:
            self._launch_input_thread()

    def add_pipes(self, input=None, output=None):
        """Add input/output pipes."""
        self._add_input_streams(*_make_iterable(input))
        self._add_output_streams(*_make_iterable(output))

    def remove_pipes(self, input=None, output=None):
        """Remove input/output pipes."""
        self._remove_input_streams(*_make_iterable(input))
        self._remove_output_streams(*_make_iterable(output))

    def toggle_output_stream(self, stream):
        """Toggle copying of all screen I/O to stream."""
        stream = self._get_wrapped_output_stream(stream)
        if stream.name in (_stream.name for _stream in self._output_streams):
            self._output_streams.remove(stream)
        else:
            self._output_streams.append(stream)

    def _add_input_streams(self, *input_streams):
        """Attach input streams."""
        if not input_streams:
            return
        first_streams = not self._input_streams
        has_stdin = any(_stream.name == stdio.stdin.name for _stream in self._input_streams)
        for stream in input_streams:
            stream = self._get_wrapped_input_stream(stream)
            if not (has_stdin and stream.name == stdio.stdin.name):
                # include stdin stream at most once, others may be replicated
                self._input_streams.append(stream)
        if first_streams and self._input_streams:
            self._launch_input_thread()

    def _remove_input_streams(self, *input_streams):
        """Detach output streams."""
        for stream_to_remove in input_streams:
            name = self._get_wrapped_output_stream(stream_to_remove).name
            for stream in self._input_streams:
                # remove the first stream whose name matches
                if stream.name == name:
                    self._input_streams.remove(stream)
                    break
            else:
                raise ValueError("can't remove input stream {}, not attached".format(name))

    def _get_wrapped_input_stream(self, stream):
        """Interpret stream argument and get the appropriate stream."""
        if stream in (u'stdio', b'stdio'):
            # sentinel value
            # stdin needs to be picked at runtime as both console.stdin and sys.stdin can change
            # and we would be keeping a reference to a closed/bad file
            stream = stdio.stdin
        elif not hasattr(stream, 'read'):
            raise TypeError(
                'input_streams must be file-like or "stdio", not `%s`'
                % (type(stream),)
            )
        if not hasattr(stream, 'name'):
            stream.name = 'input_' + random_id(8)
        return NonBlockingInputWrapper(
            stream, self._codepage, lfcr=not WIN32 and stream.isatty()
        )

    def _add_output_streams(self, *output_streams):
        """Attach output streams."""
        has_stdout = any(_stream.name == stdio.stdout.name for _stream in self._output_streams)
        for stream in output_streams:
            stream = self._get_wrapped_output_stream(stream)
            # include stdout stream at most once, others may be replicated
            # (if you do need multiple stdio streams, change the stream name)
            # this avoids duplicating stdio steams on resume
            if not (has_stdout and stream.name == stdio.stdout.name):
                self._output_streams.append(stream)

    def _remove_output_streams(self, *output_streams):
        """Detach output streams."""
        for stream_to_remove in output_streams:
            name = self._get_wrapped_output_stream(stream_to_remove).name
            for stream in self._output_streams:
                # remove the first stream whose name matches
                if stream.name == name:
                    self._output_streams.remove(stream)
                    break
            else:
                raise ValueError("can't remove output stream {}, not attached".format(name))

    def _get_wrapped_output_stream(self, stream):
        """Interpret stream argument and get the appropriate stream."""
        if stream in (u'stdio', b'stdio'):
            stream = stdio.stdout
        elif not hasattr(stream, 'write'):
            raise TypeError(
                'output_streams must be file-like or "stdio", not `%s`'
                % (type(stream),)
            )
        if not hasattr(stream, 'name'):
            stream.name = 'output_' + random_id(8)
        return self._codepage.wrap_output_stream(stream, preserve=CONTROL)

    def close(self):
        """Kill threads before exit."""
        self._stop_threads = True

    def flush(self):
        """Flush output streams."""
        for f in self._output_streams:
            f.flush()

    def write(self, s):
        """Write bytes to all stream outputs."""
        for f in self._output_streams:
            f.write(s)

    @contextmanager
    def activate(self):
        """Grab and release input stream."""
        self._active = True
        try:
            yield
        finally:
            self._active = False


    def _launch_input_thread(self):
        """Launch a thread to allow nonblocking reads on both Windows and Unix."""
        thread = threading.Thread(target=self._process_input, args=())
        thread.daemon = True
        thread.start()

    def _process_input(self):
        """Process input from streams."""
        while True:
            time.sleep(TICK)
            if self._stop_threads:
                return
            if not self._active:
                continue
            for stream in self._input_streams:
                instr = stream.read()
                if instr is None:
                    self._remove_closed_stream(stream)
                elif instr:
                    self._queues.inputs.put(signals.Event(signals.STREAM_CHAR, (instr,)))

    def _remove_closed_stream(self, stream):
        """
        Remove a closed stream from the list.
        """
        if len(self._input_streams) == 1:
            # exit the interpreter instead of closing last input
            # the input is preserved for resume
            self._queues.inputs.put(signals.Event(signals.STREAM_CLOSED))
        else:
            # input stream is closed, remove it
            self._input_streams.remove(stream)


def _make_iterable(arg):
    """Make the argument iterable, don't iterate over files or strings."""
    if not arg:
        return ()
    if hasattr(arg, 'read') or isinstance(arg, (bytes, text_type)):
        return (arg,)
    try:
        iter(arg)
    except TypeError:
        return (arg,)
    return arg



class NonBlockingInputWrapper(object):
    """
    Non-blocking input wrapper, converts CRLF.
    Wraps unicode or bytes stream; always produces unicode.
    """

    def __init__(self, stream, codepage, lfcr):
        """Set up codec."""
        self._stream = stream
        self._lfcr = lfcr
        # codepage, used to read unicode from bytes streams
        self._codepage = codepage
        try:
            self.name = stream.name
        except AttributeError:
            self.name = ''

    def read(self):
        """Read all chars available; nonblocking; returns unicode."""
        # we need non-blocking readers to be able to meaningfully have multiple inputs
        # this way we can have multiple files being read in order provided
        # while also reading from e.g. stdin on an interactive basis
        # it also enables us to convert bytes to unicode
        # assuming multibyte code sequences are read in one go.
        s = read_all_available(self._stream)
        # can be None (closed) or b'' (no input)
        if s is None:
            return None
        elif not s:
            return u''
        if isinstance(s, bytes):
            # raw input means it's already in the BASIC codepage
            # but the keyboard functions use unicode
            # for input, don't use lead-byte buffering beyond the convert call
            s = self._codepage.bytes_to_unicode(s, preserve=CONTROL)
        # replace CRLF (and, if desired, LF) with CR
        s = s.replace(u'\r\n', u'\r')
        if self._lfcr:
            s = s.replace(u'\n', u'\r')
        return s



"""
PC-BASIC - clock.py
Time and date utilities

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import datetime

from .base import error
from . import values


class Clock(object):

    def __init__(self, values):
        """Initialise clock."""
        # datetime offset for duration of the run
        # (so that we don't need permission to touch the system clock)
        # given in seconds
        self._values = values
        self.time_offset = datetime.timedelta()

    def get_time_ms(self):
        """Get milliseconds since midnight."""
        now = datetime.datetime.now() + self.time_offset
        midnight = datetime.datetime(now.year, now.month, now.day)
        diff = now-midnight
        seconds = diff.seconds
        micro = diff.microseconds
        return int(seconds*1000 + micro/1000)

    def timer_(self, args):
        """TIMER: get clock ticks since midnight."""
        list(args)
        # precision of GWBASIC TIMER is about 1/20 of a second
        timer = float(self.get_time_ms()//50) / 20.
        return self._values.new_single().from_value(timer)

    def time_(self, args):
        """TIME: Set the system time offset."""
        timestr = values.next_string(args)
        list(args)
        # allowed formats:  hh   hh:mm   hh:mm:ss  where hh 0-23, mm 0-59, ss 0-59
        now = datetime.datetime.now() + self.time_offset
        strlist = timestr.replace(b'.', b':').split(b':')
        if len(strlist) == 1:
            strlist = strlist[0].split(b'.')
        if len(strlist) not in (1, 2, 3):
            raise error.BASICError(error.IFC)
        try:
            timelist = [int(s) for s in strlist]
        except ValueError:
            raise error.BASICError(error.IFC)
        timelist += [0] * (3 - len(timelist))
        if timelist[0] > 23 or timelist[1] > 59 or timelist[2] > 59:
            raise error.BASICError(error.IFC)
        newtime = datetime.datetime(now.year, now.month, now.day,
                    timelist[0], timelist[1], timelist[2], now.microsecond)
        self.time_offset += newtime - now

    def date_(self, args):
        """DATE: Set the system date offset."""
        datestr = values.next_string(args)
        # allowed formats:
        # mm/dd/yy  or mm-dd-yy  mm 0--12 dd 0--31 yy 80--00--77
        # mm/dd/yyyy  or mm-dd-yyyy  yyyy 1980--2099
        now = datetime.datetime.now() + self.time_offset
        strlist = datestr.replace(b'/', b'-').split(b'-')
        if len(strlist) != 3:
            raise error.BASICError(error.IFC)
        try:
            datelist = [int(s) for s in strlist]
        except ValueError:
            raise error.BASICError(error.IFC)
        if (
                datelist[0] > 12 or datelist[1] > 31 or
                (datelist[2] > 77 and datelist[2] < 80) or
                (datelist[2] > 99 and datelist[2] < 1980 or datelist[2] > 2099)
            ):
            raise error.BASICError(error.IFC)
        if datelist[2] <= 77:
            datelist[2] = 2000 + datelist[2]
        elif datelist[2] < 100 and datelist[2] > 79:
            datelist[2] = 1900 + datelist[2]
        try:
            newtime = datetime.datetime(
                datelist[2], datelist[0], datelist[1],
                now.hour, now.minute, now.second, now.microsecond
            )
        except ValueError:
            raise error.BASICError(error.IFC)
        list(args)
        self.time_offset += newtime - now

    def time_fn_(self, args):
        """Get (offset) system time."""
        list(args)
        time = (datetime.datetime.now() + self.time_offset).strftime('%H:%M:%S')
        return self._values.new_string().from_str(time.encode('ascii'))

    def date_fn_(self, args):
        """Get (offset) system date."""
        list(args)
        date = (datetime.datetime.now() + self.time_offset).strftime('%m-%d-%Y')
        return self._values.new_string().from_str(date.encode('ascii'))



"""
PC-BASIC - GW-BASIC/BASICA/Cartridge BASIC compatible interpreter

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import io
import os
import sys
import locale
import logging
import traceback

from . import config
from . import info
from .basic import Session
from .debug import DebugSession
from .guard import ExceptionGuard
from .basic import NAME, VERSION, LONG_VERSION, COPYRIGHT
from .interface import Interface, InitFailed
from .compat import stdio, resources, nullcontext
from .compat import script_entry_point_guard


def main(*arguments):
    """Initialise, parse arguments and perform requested operations."""
    with config.TemporaryDirectory(prefix='pcbasic-') as temp_dir:
        # get settings and prepare logging
        settings = config.Settings(temp_dir, arguments)
        if settings.version:
            # print version and exit
            _show_version(settings)
        elif settings.help:
            # print usage and exit
            _show_usage()
        elif settings.convert:
            # convert and exit
            _convert(settings)
        elif settings.interface:
            # start an interpreter session with interface
            _run_session_with_interface(settings)
        else:
            # start an interpreter session with standard i/o
            _run_session(**settings.launch_params)


def _show_usage():
    """Show usage description."""
    usage = resources.read_text(__package__ + '.' + 'data', 'USAGE.txt', errors='replace')
    stdio.stdout.write(usage)

def _show_version(settings):
    """Show version with optional debugging details."""
    if settings.debug:
        stdio.stdout.write(info.get_version_info())
        stdio.stdout.write(info.get_platform_info())
    else:
        stdio.stdout.write(u'%s %s\n%s\n' % (NAME, VERSION, COPYRIGHT))

def _convert(settings):
    """Perform file format conversion."""
    mode, in_name, out_name = settings.conv_params
    with Session(**settings.session_params) as session:
        # binary stdin if no name supplied - use BytesIO buffer for seekability
        with session.bind_file(in_name or io.BytesIO(stdio.stdin.buffer.read())) as infile:
            session.execute(b'LOAD "%s"' % (infile,))
        with session.bind_file(out_name or stdio.stdout.buffer, create=True) as outfile:
            mode_suffix = b',%s' % (mode.encode('ascii'),) if mode.upper() in ('A', 'P') else b''
            session.execute(b'SAVE "%s"%s' % (outfile, mode_suffix))


def _run_session_with_interface(settings):
    """Start an interactive interpreter session."""
    try:
        interface = Interface(**settings.iface_params)
    except InitFailed as e: # pragma: no cover
        logging.error(e)
    else:
        exception_guard = ExceptionGuard(interface, **settings.guard_params)
        interface.launch(
            _run_session,
            interface=interface,
            exception_handler=exception_guard,
            **settings.launch_params
        )

def _run_session(
        interface=None, exception_handler=nullcontext,
        resume=False, debug=False, state_file=None,
        prog=None, commands=(), keys=u'', greeting=True, **session_params
    ):
    """Start or resume session, handle exceptions, suspend on exit."""
    if resume:
        try:
            session = Session.resume(state_file)
            session.add_pipes(**session_params)
        except Exception as e:
            # if we were told to resume but can't, give up
            logging.critical('Failed to resume session from %s: %s' % (state_file, e))
            sys.exit(1)
    elif debug:
        session = DebugSession(**session_params)
        exception_handler = nullcontext
    else:
        session = Session(**session_params)
    with exception_handler(session) as handler:
        with session:
            try:
                _operate_session(session, interface, prog, commands, keys, greeting)
            finally:
                try:
                    session.suspend(state_file)
                except Exception as e:
                    logging.error('Failed to save session to %s: %s', state_file, e)
    if exception_handler is not nullcontext and handler.exception_handled:
        _run_session(
            interface, exception_handler, resume=True, state_file=state_file, greeting=False
        )


def _operate_session(session, interface, prog, commands, keys, greeting):
    """Run an interactive BASIC session."""
    session.attach(interface)
    if greeting:
        session.greet()
    if prog:
        with session.bind_file(prog) as progfile:
            session.execute(b'LOAD "%s"' % (progfile,))
    session.press_keys(keys)
    for cmd in commands:
        session.execute(cmd)
    session.interact()



"""
PC-BASIC - GW-BASIC/BASICA/Cartridge BASIC compatible interpreter

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from .main import main, script_entry_point_guard

with script_entry_point_guard():
    main()



"""
PC-BASIC - compat.python3
Python 2/Python 3 functionality

Contains lines of code from package six, which is
Copyright (c) 2010-2018 Benjamin Peterson
and released under an MIT licence https://opensource.org/licenses/MIT
"""

import sys
import os
import struct


# deal with broken pipes

def is_broken_pipe(e):
    return isinstance(e, BrokenPipeError)

# strings

int2byte = struct.Struct(">B").pack

def add_str(cls):
    """Decorator to implement the correct str() function."""
    try:
        cls.__str__ = cls.__unicode__
    except AttributeError:
        pass
    return cls

# unicode system interfaces

getcwdu = os.getcwd
getenvu = os.getenv
iterenvu = os.environ.keys

def setenvu(key, value):
    os.environ[key] = value

# iterators
zip = zip
xrange = range

def iterchar(s):
    """Iterate over bytes, returning char."""
    return (s[_i:_i+1] for _i in range(len(s)))

def iterbytes(s):
    """Iterate over bytes/bytearray/memoryview, returning int."""
    return s

def iteritems(d, **kw):
    return iter(d.items(**kw))

def itervalues(d, **kw):
    return iter(d.values(**kw))

def iterkeys(d, **kw):
    return iter(d.keys(**kw))



"""
PC-BASIC - compat.posix
Interface for Unix-like system calls

(c) 2018--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

# pylint: disable=no-name-in-module

import os
import sys
import locale
import subprocess
import array


from .base import PY2, HOME_DIR, MACOS

if PY2: # pragma: no cover
    from .python2 import which
else:
    from shutil import which


# text conventions
# ctrl+D
EOF = u'\x04'
# LF end-of-line
EOL = u'\n'

# shell conventions
# console encoding
SHELL_ENCODING = sys.stdin.encoding or locale.getpreferredencoding()
OEM_ENCODING = SHELL_ENCODING
FS_ENCODING = sys.getfilesystemencoding()
# window suppression not needed on Unix
HIDE_WINDOW = None


# output buffer for ioctl call
_sock_size = array.array('i', [0])


##############################################################################
# various

def set_dpi_aware():
    """Enable HiDPI awareness."""

##############################################################################
# file system

def get_free_bytes(path):
    """Return the number of free bytes on the drive."""
    # note that from Python 3.3 onwards, we can use the cross-platform shutil.disk_usage
    st = os.statvfs(path)
    return st.f_bavail * st.f_frsize

def get_short_pathname(native_path):
    """Return Windows short path name or None if not available."""
    return None

if PY2: # pragma: no cover
    # the official parameter should be LC_CTYPE but that's None in my locale
    # on Windows, this would only work if the mbcs CP_ACP includes the characters we need;
    argv = [_arg.decode(FS_ENCODING, errors='replace') for _arg in sys.argv]

def is_hidden(path):
    """File is hidden."""
    # dot files are hidden on unixy systems
    base = os.path.basename(path)
    return base.startswith(u'.') and (base not in (u'.', u'..'))


##############################################################################
# printing

def line_print(printbuf, printer):
    """Print the buffer to a Unix printer, using PAPS if available."""
    if not printbuf:
        return
    if which('paps'):
        # PAPS does not recognise CRLF
        printbuf = printbuf.replace(b'\r\n', b'\n')
        # A4 paper is 595 points wide by 842 points high.
        # Letter paper is 612 by 792 points.
        # the below seems to allow 82 chars horizontally on A4; it appears
        # my PAPS version doesn't quite use cpi correctly as 10cpi should
        # allow 80 chars on A4 with a narrow margin but only does so with a
        # margin of 0.
        paps = subprocess.Popen((
                b'paps', b'--cpi=11', b'--lpi=6',
                b'--left-margin=20', b'--right-margin=20',
                b'--top-margin=6', b'--bottom-margin=6'
            ),
            stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        lpr_stdin = paps.stdout
    else:
        paps = None
        lpr_stdin = subprocess.PIPE
    command = [b'lpr']
    if printer and printer != u'default':
        command += [b'-P', printer.encode(SHELL_ENCODING, 'replace')]
    lpr = subprocess.Popen(command, stdin=lpr_stdin)
    if paps:
        proc = paps
    else:
        proc = lpr
    proc.stdin.write(printbuf)
    proc.stdin.close()



"""
PC-BASIC - randomiser.py
Random number generator

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct
from . import values

from ...compat import int2byte


class Randomiser(object):
    """Linear Congruential Generator """

    _step = 4455680 # 0x43fd00
    _period = 2**24
    _multiplier = 214013
    _increment = 2531011

    def __init__(self, values):
        """Initialise the random number generator."""
        self.clear()
        self._values = values

    def clear(self):
        """Reset the random number generator."""
        self._seed = 5228370 # 0x4fc752

    def reseed(self, val):
        """Reseed the random number generator."""
        # RANDOMIZE converts to int in a non-standard way -
        # looking at the first two bytes in the internal representation
        # on a program line, if a number outside the signed int range (or -32768) is entered,
        # the number is stored as a MBF double or float. Randomize then:
        #   - ignores the first 4 bytes (if it's a double)
        #   - reads the next two
        #   - xors them with the final two (most significant including sign bit, and exponent)
        # and interprets them as a signed int
        # e.g. 1#    = /x00/x00/x00/x00 /x00/x00/x00/x81 gets read as /x00/x00 ^ /x00/x81
        #            = /x00/x81 -> 0x10000-0x8100 = -32512 (sign bit set)
        #      0.25# = /x00/x00/x00/x00 /x00/x00/x00/x7f gets read as /x00/x00 ^ /x00/x7f
        #            = /x00/x7F -> 0x7F00 = 32512 (sign bit not set)
        #              /xDE/xAD/xBE/xEF /xFF/x80/x00/x80 gets read as /xFF/x80 ^ /x00/x80
        #            = /xFF/x00 -> 0x00FF = 255
        s = val.to_bytes()
        final_two = s[-2:]
        mask = bytearray(2)
        if len(s) >= 4:
            mask = s[-4:-2]
        final_two = int2byte(final_two[0]^mask[0]) + int2byte(final_two[1]^mask[1])
        # unpack to signed integer
        n, = struct.unpack('<h', final_two)
        self._seed &= 0xff
        self._cycle()
        self._seed += n * self._step
        self._seed %= self._period

    def rnd_(self, args):
        """Get a value from the random number generator."""
        f, = args
        if f is None:
            # RND
            self._cycle()
        else:
            f = values.to_single(f)
            if f.is_zero():
                # RND(0) returns last value
                pass
            else:
                if f.is_negative():
                    # re-seed with integer value of mantissa
                    # so RND(-1) does not mean n==-1 but n==-2**23
                    self._seed = -f.mantissa()
                self._cycle()
        # seed/period
        return self._values.new_single().from_int(self._seed).idiv(
            self._values.new_single().from_int(self._period)
        )

    def _cycle(self):
        """Move the random number generator to the next state."""
        self._seed = (self._seed*self._multiplier + self._increment) % self._period



"""
PC-BASIC - codepage.py
Codepage conversions

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import unicodedata
import logging
import codecs
import os
import io

from ..compat import iterchar, iteritems, int2byte, unichr
from ..compat import StreamWrapper, is_readable_text_stream, is_writable_text_stream
from .data import DEFAULT_CODEPAGE

# characters in the printable ASCII range 0x20-0x7E cannot be redefined
# but can have their glyphs subsituted - they will work and transcode as the
# ASCII but show as the subsitute glyph. Used e.g. for YEN SIGN in Shift-JIS
# see http://www.siao2.com/2005/09/17/469941.aspx
PRINTABLE_ASCII = tuple(int2byte(_c) for _c in range(0x20, 0x7F))

# on the terminal, these values are not shown as special graphic chars but as their normal effect
# BEL, TAB, LF, HOME, CLS, CR, RIGHT, LEFT, UP, DOWN  (and not BACKSPACE)
CONTROL = (b'\x07', b'\x09', b'\x0A', b'\x0B', b'\x0C', b'\x0D', b'\x1C', b'\x1D', b'\x1E', b'\x1F')

# left-connecting box drawing chars for two box-character sets (single line, double line)
_BOX_LEFT_UNICODE = (u'\u2500', u'\u2550')

# right-connecting box drawing chars (single line, double line)
_BOX_RIGHT_UNICODE = (u'\u2500', u'\u2550')


# all box-drawing chars
# these were not actually supported by mbcs TSRs and aren't used by PC-BASIC
# but kept here for reference

# left-connecting
## single line:
# 0x2500, 0x252c, 0x2534, 0x253c, 0x2510, 0x2518, 0x2524,
# mixed single / double
# 0x2556, 0x255c, 0x2562, 0x2565, 0x2568, 0x256b, 0x256e,
# dotted lines
# 0x2504, 0x2508, 0x254c,
# mixed thick / thin line
# 0x251a, 0x2526, 0x2527, 0x2528, 0x252e, 0x2530, 0x2532, 0x2536, 0x2538, 0x253a, 0x253e,
# 0x2540, 0x2541, 0x2542, 0x2544, 0x2546, 0x254a, 0x257c
# rounded corners and half-lines
# 0x256f, 0x2574,
## double line:
# 0x2550, 0x2566, 0x2569, 0x256c,0x2557, 0x255d, 0x2563,
# mixed single / double line
# 0x2555, 0x255b, 0x2561, 0x2564, 0x2567, 0x256a,

# right-connecting
# single line
## 0x2500, 0x252c, 0x2534, 0x253c,
## 0x250c, 0x2514, 0x251c,
# dotted
# 0x2504, 0x2508,
# mixed
# 0x2516, 0x251e, 0x251f, 0x2520, 0x252d,
# 0x2530, 0x2531, 0x2535, 0x2538, 0x2539, 0x253d,
# 0x2540, 0x2541, 0x2542, 0x2543, 0x2545, 0x2549, 0x254c,
# 0x2553, 0x2559, 0x255f, 0x2565, 0x2568, 0x256b, 0x256d, 0x2570, 0x2576, 0x257e
# double line
## 0x2550, 0x2566, 0x2569, 0x256c,
## 0x2554, 0x255a, 0x2560,
# 0x2552, 0x2558, 0x255e, 0x2564, 0x2567, 0x256a,


###############################################################################
# codepages

class Codepage(object):
    """Codepage tables."""

    def __init__(self, codepage_dict=None, box_protect=True):
        """Load and initialise codepage tables."""
        codepage_dict = codepage_dict or {
            int2byte(_i): _c
            for _i, _c in enumerate(DEFAULT_CODEPAGE)
        }
        # protect box drawing sequences under dbcs?
        self.box_protect = box_protect
        # lead and trail bytes
        self.lead = set()
        self.trail = set()
        # box-protection sets
        self._box_left = [set(), set()]
        self._box_right = [set(), set()]
        # main dictionary
        self._cp_to_unicode = {}
        # glyph subsitutes for printable ascii
        self._substitutes = {}
        dbcs_num_chars = 0
        for cp_point, unicode_cluster in iteritems(codepage_dict):
            # normalise clusters so we can match later
            unicode_cluster = unicodedata.normalize('NFC', unicode_cluster)
            # do not redefine printable ASCII, but substitute glyphs
            if (
                    cp_point in PRINTABLE_ASCII and
                    (len(unicode_cluster) > 1 or ord(unicode_cluster) != ord(cp_point))
                ):
                ascii_cp = unichr(ord(cp_point))
                self._cp_to_unicode[cp_point] = ascii_cp
                self._substitutes[cp_point] = unicode_cluster
            else:
                self._cp_to_unicode[cp_point] = unicode_cluster
            # track lead and trail bytes
            if len(cp_point) == 2:
                self.lead.add(cp_point[0:1])
                self.trail.add(cp_point[1:2])
                dbcs_num_chars += 1
            # track box drawing chars
            else:
                for i in (0, 1):
                    if unicode_cluster in _BOX_LEFT_UNICODE[i]:
                        self._box_left[i].add(cp_point[0:1])
                    if unicode_cluster in _BOX_RIGHT_UNICODE[i]:
                        self._box_right[i].add(cp_point[0:1])
        # fill up any undefined 1-byte codepoints
        for c in range(256):
            if int2byte(c) not in self._cp_to_unicode:
                self._cp_to_unicode[int2byte(c)] = u'\0'
        self._unicode_to_cp = dict((reversed(_item) for _item in iteritems(self._cp_to_unicode)))
        self._inverse_substitutes = dict((reversed(_item) for _item in iteritems(self._substitutes)))
        # keep set of clusters of more than one unicode code point
        self._unicode_clusters = set(
            _cluster for _cluster in self._unicode_to_cp if len(_cluster) > 1
        )
        # ensure longest sequences get checked first (greedy clustering)
        self._unicode_clusters = list(reversed(sorted(self._unicode_clusters, key=len)))
        # is the current codepage a double-byte codepage?
        self.dbcs = dbcs_num_chars > 0

    def connects(self, c, d, bset):
        """Return True if c and d connect according to box-drawing set bset."""
        return c in self._box_right[bset] and d in self._box_left[bset]

    def _from_unicode(self, uc, errors='ignore'):
        """Convert NFC unicode cluster to codepage char sequence."""
        # pass through eascii clusters
        if uc.startswith(u'\0'):
            return b''.join(int2byte(min(255, ord(c))) for c in uc)
        try:
            return self._inverse_substitutes[uc]
        except KeyError:
            pass
        try:
            # try to codepage-encode the unicode char
            return self._unicode_to_cp[uc]
        except KeyError:
            # pass control sequences as ascii. this includes \r.
            # control sequences are not in the dictionary
            # because it holds their graphical replacement characters.
            # ignore/raise/replace everything else (unicode chars not in codepage)
            return uc.encode('ascii', errors=errors)

    def _split_unicode(self, ucs):
        """Split unicode string into codepage-dependent clusters; preserve e-ascii clusters."""
        # clusters are stored as C normal form (combine what can be combined)
        ucs = unicodedata.normalize('NFC', ucs)
        clusters = []
        while ucs:
            if ucs[0] == u'\0' and ucs[1:2] and ord(ucs[1:2]) < 256:
                # preserve e-ascii clusters
                length = 2
            else:
                # if a cluster is matched, preserve that; otherwise, use single unicode point
                # i.e. if a perceived grapheme cluster is not contained in the codepage
                # it will be passed codepoint-by-codepoint (and likely ignored/raised/replaced)
                length = 1
                for cluster in self._unicode_clusters:
                    if ucs.startswith(cluster):
                        length = len(cluster)
                        break
            clusters.append(ucs[:length])
            ucs = ucs[length:]
        return clusters

    def unicode_to_bytes(self, ucs, errors='ignore'):
        """Convert unicode string to codepage string."""
        return b''.join(self._from_unicode(uc, errors=errors) for uc in self._split_unicode(ucs))

    def codepoint_to_unicode(self, cp, replace=u'', use_substitutes=False):
        """Convert codepage point to unicode grapheme cluster."""
        if use_substitutes and self._substitutes:
            try:
                return self._substitutes[cp]
            except KeyError:
                pass
        return self._cp_to_unicode.get(cp, replace)

    def bytes_to_unicode(self, cps, preserve=(), box_protect=None, use_substitutes=False):
        """Convert codepage string to unicode string."""
        if box_protect is None:
            box_protext = self.box_protect
        return Converter(self, preserve, box_protect, use_substitutes).to_unicode(cps, flush=True)

    def get_converter(self, preserve=(), use_substitutes=False):
        """Get converter from codepage to unicode."""
        return Converter(self, preserve, self.box_protect, use_substitutes=use_substitutes)

    def wrap_output_stream(self, stream, preserve=()):
        """Wrap a stream so that we can write codepage bytes to it."""
        # check for file-like objects that expect unicode, raw output otherwise
        if not is_writable_text_stream(stream):
            return stream
        return OutputStreamWrapper(stream, self, preserve)

    def wrap_input_stream(self, stream, replace_newlines=False):
        """Wrap a stream so that we can read codepage bytes from it."""
        # check for file-like objects that expect unicode, raw output otherwise
        if is_readable_text_stream(stream):
            stream = InputStreamWrapper(stream, self)
        if replace_newlines:
            return NewlineWrapper(stream)
        return stream


##############################################################################
# stream wrappers


class OutputStreamWrapper(StreamWrapper):
    """
    Converter stream wrapper, takes bytes input.
    Stream must be a unicode (text) stream.
    """

    def __init__(self, stream, codepage, preserve=()):
        """Set up codec."""
        self._conv = codepage.get_converter(preserve)
        self._stream = stream

    def write(self, s):
        """Write bytes to codec stream."""
        # decode BASIC bytes --(codepage)-> unicode
        self._stream.write(self._conv.to_unicode(s))


class InputStreamWrapper(StreamWrapper):
    """
    Converter stream wrapper, produces bytes output.
    Stream must be a unicode (text) stream.
    """

    def __init__(self, stream, codepage):
        """Set up codec."""
        self._codepage = codepage
        self._stream = stream
        self._buffer = b''

    def read(self, n=-1):
        """Read n bytes from stream with codepage conversion."""
        if n > len(self._buffer):
            unistr = self._stream.read(n - len(self._buffer))
        elif n == -1:
            unistr = self._stream.read()
        else:
            unistr = u''
        converted = (self._buffer + self._codepage.unicode_to_bytes(unistr, errors='replace'))
        if n < 0:
            output = converted
        else:
            output, self._buffer = converted[:n], converted[n:]
        return output


class NewlineWrapper(StreamWrapper):
    """Replace newlines on input stream. Wraps a bytes stream."""

    def __init__(self, stream):
        """Set up codec."""
        StreamWrapper.__init__(self, stream)
        self._last = b''

    def read(self, n=-1):
        """Read n bytes from stream with codepage conversion."""
        if n == 0:
            return b''
        output = b''
        while n < 0 or len(output) < n:
            new_bytes = self._stream.read(n - len(output))
            # empty means end of file
            if not new_bytes:
                break
            last, self._last = self._last, new_bytes[-1:]
            # absorb CR LF
            if last == b'\r' and new_bytes[:1] == b'\n':
                new_bytes = new_bytes[1:]
            output += new_bytes
        output = output.replace(b'\n', b'\r')
        return output


##################################################
# conversion with box protection

class Converter(object):
    """Buffered converter to Unicode - supports DBCS and box-drawing protection."""

    def __init__(self, codepage, preserve=(), box_protect=None, use_substitutes=False):
        """Initialise with empty buffer."""
        self._cp = codepage
        # hold one or two bytes
        # lead byte without trail byte, or box-protectable dbcs
        self._buf = b''
        # preserve is a tuple/list of bytes that should keep the same ordinal
        # this is mainly for control characters that have alternate graphical symbols
        self._preserve = set(preserve)
        # may override box protection defaults
        self._box_protect = box_protect or self._cp.box_protect
        self._use_substitutes = use_substitutes
        self._dbcs = self._cp.dbcs
        self._bset = -1
        self._last = b''

    def to_unicode(self, s, flush=False):
        """Process codepage string, returning unicode string when ready."""
        return u''.join(self.to_unicode_list(s, flush))

    def to_unicode_list(self, s, flush=False):
        """Convert codepage to list of unicode with fullwidth marked by trailing u''."""
        tuples = ((_seq,) if len(_seq) == 1 else (_seq, b'') for _seq in self._mark(s, flush))
        sequences = (_seq for _tup in tuples for _seq in _tup)
        return [
            (
                _seq.decode('ascii', errors='ignore')
                if (_seq in self._preserve)
                else self._cp.codepoint_to_unicode(_seq, use_substitutes=self._use_substitutes)
            )
            for _seq in sequences
        ]

    def _mark(self, s, flush=False):
        """Convert bytes to list of codepage bytes."""
        if not self._dbcs:
            # stateless if not dbcs
            return list(iterchar(s))
        else:
            sequences = [seq for c in iterchar(s) for seq in self._process(c)]
            if flush:
                sequences += self._flush()
            return sequences

    def _flush(self, num=None):
        """Empty buffer and return contents."""
        out = []
        if num is None:
            num = len(self._buf)
        if self._buf:
            # can be one or two-byte sequence in self._buf
            out.append(self._buf[:num])
        self._buf = self._buf[num:]
        return out

    def _process(self, c):
        """Process a single char, returning unicode char sequences when ready """
        if not self._box_protect:
            return self._process_nobox(c)
        out = []
        if c in self._preserve:
            # control char; flush buffer as SBCS and add control char unchanged
            out += self._flush() + [c]
            self._bset = -1
            self._last = b''
        elif self._bset == -1:
            if not self._buf:
                out += self._process_case0(c)
            elif len(self._buf) == 1:
                out += self._process_case1(c)
            elif len(self._buf) == 2:
                out += self._process_case2(c)
            else: # pragma: no cover
                # not allowed
                logging.debug(b'DBCS buffer corrupted: %d %s', self._bset, repr(self._buf))
        elif len(self._buf) == 2:
            out += self._process_case3(c)
        elif not self._buf:
            out += self._process_case4(c)
        else: # pragma: no cover
            # not allowed
            logging.debug(b'DBCS buffer corrupted: %d %s', self._bset, repr(self._buf))
        return out

    def _process_nobox(self, c):
        """Process a single char, no box drawing protection """
        out = []
        if c in self._preserve:
            # control char; flush buffer as SBCS and add control char unchanged
            out += self._flush() + [c]
            return out
        elif self._buf:
            if c in self._cp.trail:
                # add a DBCS character
                self._buf += c
                out += self._flush()
                return out
            else:
                # flush buffer
                out += self._flush()
        if c in self._cp.lead:
            self._buf = c
        else:
            out.append(c)
        return out

    def _process_case0(self, c):
        """Process a single char with box drawing protection; case 0, starting point """
        out = []
        if c not in self._cp.lead:
            out.append(c)
            # goes to case 0
        else:
            self._buf += c
            # goes to case 1
        return out

    def _process_case1(self, c):
        """Process a single char with box drawing protection; case 1 """
        out = []
        if c not in self._cp.trail:
            out += self._flush() + [c]
            # goes to case 0
        else:
            for bset in (0, 1):
                if self._cp.connects(self._buf, c, bset):
                    self._bset = bset
                    self._buf += c
                    break
                    # goes to case 3
            else:
                # no connection
                self._buf += c
                # goes to case 2
        return out

    def _process_case2(self, c):
        """Process a single char with box drawing protection; case 2 """
        out = []
        if c not in self._cp.lead:
            out += self._flush() + [c]
            # goes to case 0
        else:
            for bset in (0, 1):
                if self._cp.connects(self._buf[-1:], c, bset):
                    self._bset = bset
                    # take out only first byte
                    out += self._flush(1)
                    self._buf += c
                    break
                    # goes to case 3
            else:
                # no connection found
                out += self._flush()
                self._buf += c
                # goes to case 1
        return out

    def _process_case3(self, c):
        """Process a single char with box drawing protection; case 3 """
        out = []
        if c not in self._cp.lead:
            out += self._flush() + [c]
        elif self._cp.connects(self._buf[-1:], c, self._bset):
            self._last = self._buf[-1:]
            # output box drawing
            out += self._flush(1) + self._flush(1) + [c]
            # goes to case 4
        else:
            out += self._flush()
            self._buf = c
            self._bset = -1
            # goes to case 1
        return out

    def _process_case4(self, c):
        """Process a single char with box drawing protection; case 4, continuing box drawing """
        out = []
        if c not in self._cp.lead:
            out.append(c)
            # goes to case 0
        elif self._cp.connects(self._last, c, self._bset):
            self._last = c
            out.append(c)
            # goes to case 4
        else:
            self._buf += c
            self._bset = -1
            # goes to case 1
        return out



"""
PC-BASIC - audio_sdl2.py
Sound interface based on SDL2

(c) 2015--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

# see e.g. http://toomanyideas.net/2014/pysdl2-playing-a-sound-from-a-wav-file.html

import os
import logging
import ctypes
from collections import deque

from ..compat import zip, BASE_DIR
from .audio import AudioPlugin
from .base import audio_plugins, InitFailed
from . import synthesiser


###############################################################################
# locate and load SDL libraries

if False:
    # packagers take note, we need these imports to be bundled
    from . import sdl2loader
    from . import sdl2

# sdl2 module is imported only at plugin initialisation
sdl2 = None

# custom dll location
LIB_DIR = os.path.join(BASE_DIR, 'lib')


def _import_sdl2():
    """Import the sdl2 bindings and define constants."""
    global sdl2

    # look for SDL2.dll / libSDL2.dylib / libSDL2.so:
    # first in LIB_DIR, then in pysdl2dll module, then the standard search path
    # this means that user should not have dll in LIB_DIR if they want to use another one
    from . import sdl2loader
    sdl2loader.load_dlls(LIB_DIR)

    # this raises ImportError if no library has been found
    from . import sdl2

# approximate generator chunk length
# one wavelength at 37 Hz is 1192 samples at 44100 Hz
_CHUNK_LENGTH = 1192 * 4
# length of chunks to be consumed by callback
_CALLBACK_CHUNK_LENGTH = 2048
# number of samples below which to replenish the buffer
_MIN_SAMPLES_BUFFER = 2 * _CALLBACK_CHUNK_LENGTH
# pause audio device after given number of ticks without sound
_QUIET_QUIT = 100


@audio_plugins.register('ansi')
@audio_plugins.register('cli')
@audio_plugins.register('sdl2')
class AudioSDL2(AudioPlugin):
    """SDL2-based audio plugin."""

    def __init__(self, audio_queue, **kwargs):
        """Initialise sound system."""
        try:
            _import_sdl2()
        except ImportError:
            raise InitFailed('Module `sdl2` not found')
        # synthesisers
        self._signal_sources = synthesiser.get_signal_sources()
        # sound generators for each voice
        self._generators = [deque() for _ in synthesiser.VOICES]
        # buffer of samples; drained by callback, replenished by _play_sound
        self._samples = [bytearray() for _ in synthesiser.VOICES]
        self._next_tone = [None for _ in synthesiser.VOICES]
        self._device = None
        self._audiospec = None
        self._quiet_ticks = 0
        AudioPlugin.__init__(self, audio_queue)

    def __enter__(self):
        """Perform any necessary initialisations."""
        # init sdl audio in this thread separately
        sdl2.SDL_Init(sdl2.SDL_INIT_AUDIO)
        # SDL AudioDevice and specifications
        # S8 gives ticks on pausing and unpausing so we use S16
        audiospec = sdl2.SDL_AudioSpec(
            freq=synthesiser.SAMPLE_RATE, aformat=sdl2.AUDIO_S16LSB, channels=1,
            samples=_CALLBACK_CHUNK_LENGTH, callback=sdl2.SDL_AudioCallback(self._get_next_chunk)
        )
        self._device = sdl2.SDL_OpenAudioDevice(None, False, audiospec, None, 0)
        if self._device == 0:
            logging.warning('Could not open audio device: %s', sdl2.SDL_GetError())
        # prevent audiospec from being garbage collected as it goes out of scope
        self._audiospec = audiospec
        return AudioPlugin.__enter__(self)

    def tone(self, voice, frequency, duration, loop, volume):
        """Enqueue a tone."""
        self._generators[voice].append(synthesiser.SoundGenerator(
            self._signal_sources[voice], synthesiser.FEEDBACK_TONE,
            frequency, duration, loop, volume
        ))

    def noise(self, source, frequency, duration, loop, volume):
        """Enqueue a noise."""
        feedback = synthesiser.FEEDBACK_NOISE if source else synthesiser.FEEDBACK_PERIODIC
        self._generators[synthesiser.NOISE_VOICE].append(synthesiser.SoundGenerator(
            self._signal_sources[synthesiser.NOISE_VOICE], feedback,
            frequency, duration, loop, volume
        ))

    def hush(self):
        """Stop sound."""
        self._next_tone = [None for _ in synthesiser.VOICES]
        for gen in self._generators:
            gen.clear()
        sdl2.SDL_LockAudioDevice(self._device)
        self._samples = [bytearray() for _ in synthesiser.VOICES]
        sdl2.SDL_UnlockAudioDevice(self._device)

    def _work(self):
        """Replenish sample buffer."""
        if not any(self._generators) and not any(self._next_tone):
            if self._quiet_ticks >= _QUIET_QUIT:
                sdl2.SDL_PauseAudioDevice(self._device, 1)
            self._quiet_ticks += 1
            return
        else:
            self._quiet_ticks = 0
            sdl2.SDL_PauseAudioDevice(self._device, 0)
        for voice in synthesiser.VOICES:
            if len(self._samples[voice]) > _MIN_SAMPLES_BUFFER:
                # nothing to do
                continue
            while True:
                if self._next_tone[voice] is None or self._next_tone[voice].loop:
                    try:
                        # looping tone will be interrupted
                        # by any new tone appearing in the generator queue
                        self._next_tone[voice] = self._generators[voice].popleft()
                    except IndexError:
                        if self._next_tone[voice] is None:
                            current_chunk = None
                            break
                current_chunk = self._next_tone[voice].build_chunk(_CHUNK_LENGTH)
                if current_chunk is not None:
                    break
                self._next_tone[voice] = None
            if current_chunk is not None:
                # append chunk to samples list
                # lock to ensure callback doesn't try to access the list too
                sdl2.SDL_LockAudioDevice(self._device)
                self._samples[voice] = bytearray().join((self._samples[voice], current_chunk))
                sdl2.SDL_UnlockAudioDevice(self._device)

    def _get_next_chunk(self, notused, stream, length_bytes):
        """Callback function to generate the next chunk to be played."""
        # this assumes 8-bit samples
        length = length_bytes // 2
        # if samples have run out, add silence
        samples = (
            _samp.ljust(length, b'\0') if len(_samp) < length else _samp[:length]
            for _samp in self._samples
        )
        # mix the samples
        mixed = bytearray(sum(_b) & 0xff for _b in zip(*samples))
        # convert from S8 to S16 (little-endian)
        # interlace with leading zeros to multiply by 256
        mixed_16 = bytearray(length_bytes)
        mixed_16[1::2] = mixed
        self._samples = [_samp[length:] for _samp in self._samples]
        ctypes.memmove(
            stream, (ctypes.c_char * length_bytes).from_buffer(mixed_16), length_bytes
        )



"""
PC-BASIC - lister.py
Convert tokenised to plain-text format

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct

from ...compat import int2byte

from ..base import tokens as tk
from ..base.tokens import DIGITS, ALPHANUMERIC
from .. import values


class Lister(object):
    """BASIC detokeniser."""

    def __init__(self, values, token_dict):
        """Initialise tokeniser."""
        self._values = values
        self._token_to_keyword = token_dict.to_keyword

    def detokenise_line(self, ins):
        """Convert a tokenised program line to ascii text."""
        current_line = self.detokenise_line_number(ins)
        if current_line < 0:
            # detokenise_line_number has returned -1 and left us at: .. 00 | _00_ 00 1A
            # stream ends or end of file sequence \x00\x00\x1A
            return -1, b'', 0
        elif current_line == 0 and ins.peek() == b' ':
            # ignore up to one space after line number 0
            ins.read(1)
        linum = bytearray(b'%d' % (current_line,))
        # write one extra whitespace character after line number
        # unless first char is TAB
        if ins.peek() != b'\t':
            linum += bytearray(b' ')
        text_positions = [(ins.tell(), 0)]
        line, line_text_positions = self.detokenise_compound_statement(ins)
        line_text_positions = [
            (_bytepos, _textpos+len(linum) + 1)
            for _bytepos, _textpos in line_text_positions
        ]
        return current_line, linum + line, text_positions + line_text_positions

    def detokenise_line_number(self, ins):
        """Parse line number and leave pointer at first char of line."""
        trail = ins.read(4)
        linum = self.token_to_line_number(trail)
        # if end of program or truncated, leave pointer at start of line number C0 DE or 00 00
        if linum == -1:
            ins.seek(-len(trail), 1)
        return linum

    def token_to_line_number(self, trail):
        """Unpack a line number token trail, -1 if end of program."""
        if len(trail) < 4 or trail[:2] == b'\0\0':
            return -1
        return struct.unpack_from('<H', trail, 2)[0]

    def detokenise_compound_statement(self, ins):
        """Detokenise tokens until end of line."""
        litstring, comment = False, False
        text_positions = []
        output = bytearray()
        while True:
            s = ins.read(1)
            # keep track of matching text and binary positions
            text_positions.append((ins.tell(), len(output)))
            if s in tk.END_LINE:
                # \x00 ends lines and comments when listed,
                # if not inside a number constant
                # stream ended or end of line
                break
            elif s == b'"':
                # start of literal string, passed verbatim
                # until a closing quote or EOL comes by
                # however number codes are *printed* as the corresponding numbers,
                # even inside comments & literals
                output += s
                litstring = not litstring
            elif s in tk.NUMBER or s in tk.LINE_NUMBER:
                output += self._detokenise_number(ins, s)
            elif comment or litstring or (b'\x20' <= s <= b'\x7E'):
                # honest ASCII
                output += s
            elif s == b'\x0A':
                # LF becomes LF CR
                output += b'\x0A\x0D'
            elif s <= b'\x09':
                # controls that do not double as tokens
                output += s
            else:
                token = self._detokenise_keyword_into(ins, s, output)
                comment = token in tk.COMMENT
        return output[:255], text_positions

    def _detokenise_keyword_into(self, ins, lead, output):
        """Convert a one- or two-byte keyword token to ascii."""
        # try for single-byte token or two-byte token
        # if no match, first char is passed unchanged
        token = lead
        try:
            keyword = self._token_to_keyword[token]
        except KeyError:
            token += ins.peek()
            try:
                keyword = self._token_to_keyword[token]
                ins.read(1)
            except KeyError:
                output += token[:1]
                return False
        # letter or number followed by token is separated by a space
        if (
                token not in tk.OPERATOR
                and output and bytes(output[-1:]) in ALPHANUMERIC
                # we need to check again for FN and USR, but not SPC( and TAB(
                # because we check the converted output, not the previous token
                and not (len(output) >= 2 and bytes(output[-2:]) == tk.KW_FN)
                and not (len(output) >= 3 and bytes(output[-3:]) == tk.KW_USR)
            ):
            output += b' '
        # check for special cases
        #   [:REM']   ->  [']
        # we need to read one ahead at REM, or tk_O_REM would be transcribed as part of the comment
        # and the replacement code would not work
        next_char = ins.peek(1)
        if token == tk.REM and next_char == tk.O_REM and output and bytes(output[-1:]) == b':':
            ins.read(1)
            output[:] = output[:-1] + tk.KW_O_REM
        #   [WHILE+]  ->  [WHILE]
        elif token == tk.O_PLUS and len(output) >= 5 and bytes(output[-5:]) == tk.KW_WHILE:
            # ignore the +
            pass
        #   [:ELSE]  ->  [ELSE]
        # note that anything before ELSE gets cut off,
        # e.g. if we have 1ELSE instead of :ELSE it also becomes ELSE
        elif token == tk.ELSE:
            if not output:
                # special case at start of line, lone ELSE (not :ELSE) becomes LSE
                output += keyword[1:]
            else:
                output[:] = output[:-1] + keyword
        else:
            output += keyword
        # token followed by token or number is separated by a space,
        # except operator tokens, comment tokens and SPC(, TAB(, FN, USR
        if (
                token not in tk.OPERATOR + tk.COMMENT + (tk.TAB, tk.SPC, tk.USR, tk.FN)
                and next_char not in tk.END_LINE + tk.OPERATOR + (
                    tk.O_REM, b'"', b',', b';', b' ', b':', b'(', b')', b'$',
                    b'%', b'!', b'#', b'_', b'@', b'~', b'|', b'`'
                )
            ):
            # excluding TAB( SPC( and FN. \xD9 is ', \xD1 is FN, \xD0 is USR.
            output += b' '
        return token

    def _detokenise_number(self, ins, lead):
        """Convert number token to Python bytes string."""
        ntrail = tk.PLUS_BYTES.get(lead, 0)
        trail = ins.read(ntrail)
        if lead == tk.T_OCT:
            return b'&O' + self._values.from_bytes(trail).to_oct()
        elif lead == tk.T_HEX:
            return b'&H' + self._values.from_bytes(trail).to_hex()
        elif lead == tk.T_BYTE:
            return b'%d' % (ord(trail),)
        elif tk.C_0 <= lead <= tk.C_10:
            return b'%d' % (ord(lead) - ord(tk.C_0),)
        elif lead in tk.LINE_NUMBER:
            # 0D: line pointer (unsigned int) - this token should not be here;
            #     interpret as line number and carry on
            # 0E: line number (unsigned int)
            return b'%d' % (struct.unpack('<H', trail)[0],)
        elif lead in (tk.T_SINGLE, tk.T_DOUBLE, tk.T_INT):
            return self._values.from_bytes(trail).to_str(leading_space=False, type_sign=True)
        # failsafe line, not reached
        return b'' # pragma: no cover



"""
PC-BASIC - display package
Text and graphics buffer, cursor and screen operations

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from .display import Display
from .screencopyhandler import ScreenCopyHandler



"""
PC-BASIC - strings.py
String values

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct
import logging
from operator import itemgetter

from ...compat import iteritems

from ..base import error
from . import numbers


class String(numbers.Value):
    """String pointer."""

    sigil = b'$'
    size = 3

    def __init__(self, buffer, values):
        """Initialise the pointer."""
        numbers.Value.__init__(self, buffer, values)
        self._stringspace = values.stringspace

    def length(self):
        """String length."""
        return bytearray(self._buffer)[0]

    def address(self):
        """Pointer address."""
        return struct.unpack_from('<H', self._buffer, 1)[0]

    def dereference(self):
        """String value pointed to."""
        length, address = struct.unpack('<BH', self._buffer)
        return self._stringspace.view(length, address).tobytes()

    def from_str(self, python_str):
        """Set to value of python str."""
        assert isinstance(python_str, bytes), type(python_str)
        self._buffer[:] = struct.pack('<BH', *self._stringspace.store(python_str))
        return self

    def from_pointer(self, length, address):
        """Set buffer to string pointer."""
        self._buffer[:] = struct.pack('<BH', length, address)
        return self

    def to_pointer(self):
        """Get length and address."""
        return struct.unpack('<BH', self._buffer)

    from_value = from_str
    to_value = dereference
    to_str = dereference

    def add(self, right):
        """Concatenate strings. In-place for the pointer."""
        return self.new().from_str(self.dereference() + right.dereference())

    def eq(self, right):
        """This string equals the right-hand side."""
        return self.to_str() == right.to_str()

    def gt(self, right):
        """This string orders after the right-hand side."""
        left = self.to_str()
        right = right.to_str()
        shortest = min(len(left), len(right))
        for i in range(shortest):
            if left[i] > right[i]:
                return True
            elif left[i] < right[i]:
                return False
        # the same so far...
        # the shorter string is said to be less than the longer,
        # provided they are the same up till the length of the shorter.
        if len(left) > len(right):
            return True
        # left is shorter, or equal strings
        return False

    def lset(self, in_str, justify_right):
        """Justify a str into an existing buffer and pad with spaces."""
        # v is empty string if variable does not exist
        # trim and pad to size of target buffer
        length = self.length()
        in_str = in_str.to_value()
        if justify_right:
            in_str = in_str[:length].rjust(length)
        else:
            in_str = in_str[:length].ljust(length)
        # make a copy only if not in a writeable location
        target = self._stringspace.check_modify(*self.to_pointer())
        self.from_pointer(*target)
        # copy the new string in
        self._stringspace.view(*target)[:] = in_str
        return self

    def midset(self, start, num, val):
        """Modify a string in an existing buffer."""
        # we need to decrement basic offset by 1 to get python offset
        offset = start - 1
        # don't overwrite more of the old string than the length of the new string
        num = min(num, val.length())
        # ensure the length of source string matches target
        length = self.length()
        if offset + num > length:
            num = length - offset
        if num <= 0:
            return self
        # make a copy only if not in a writeable location
        target = self._stringspace.check_modify(*self.to_pointer())
        self.from_pointer(*target)
        source = val.to_pointer()
        if source != target:
            self._stringspace.view(*target)[offset:offset+num] = (
                self._stringspace.view(*source)[:num]
            )
        else:
            # copy byte by byte from left to right
            # to conform to GW overwriting of source string on overlap
            for i in range(num):
                self._stringspace.view(*target)[i+offset:i+offset+1] = (
                    self._stringspace.view(*source)[i:i+1]
                )
        return self


    ######################################################################
    # unary functions

    def len(self):
        """LEN: length of string."""
        return numbers.Integer(None, self._values).from_int(self.length())

    def asc(self):
        """ASC: ordinal ASCII value of a character."""
        s = bytearray(self.to_str())
        error.throw_if(not s)
        return numbers.Integer(None, self._values).from_int(s[0])

    def space(self, num):
        """SPACE$: repeat spaces."""
        num = num.to_integer().to_int()
        error.range_check(0, 255, num)
        return self.new().from_str(b' ' * num)


class StringSpace(object):
    """Table of strings accessible by their length and address."""

    def __init__(self, memory):
        """Initialise empty string space."""
        self._memory = memory
        self._strings = {}
        self._temp = None
        self.clear()

    def __repr__(self):
        """Debugging representation of string table."""
        return '\n'.join(
            '%x: %r' % (n, bytes(v))
            for n, v in iteritems(self._strings)
        )

    def clear(self):
        """Empty string space."""
        self._strings.clear()
        # strings are placed at the top of string memory, just below the stack
        self.current = self._memory.stack_start()

    def rebuild(self, stringspace):
        """Rebuild from stored copy."""
        self.clear()
        self._strings.update(stringspace._strings)
        self.current = stringspace.current

    def copy_to(self, string_space, length, address):
        """Copy a string to another string space."""
        return string_space.store(self.view(length, address).tobytes())

    def _retrieve(self, length, address):
        """Retrieve a string by its pointer."""
        # if string length == 0, return empty string
        try:
            return bytearray() if length == 0 else self._strings[address]
        except KeyError: # pragma: no cover
            raise KeyError(u'Dereferencing detached string at %x (%d)' % (address, address))

    def view(self, length, address):
        """Return a writeable view of a string from its string pointer."""
        # empty string pointers can point anywhere
        if length == 0:
            return memoryview(bytearray())
        if address >= self._memory.var_start():
            # string stored in string space
            return memoryview(self._retrieve(length, address))
        elif address >= self._memory.code_start:
            # get string stored in code as bytearray
            codestr = self._memory.program.get_memory_block(address, length)
            # NOTE this is a writeable view of a *copy* of the code!
            return memoryview(codestr)
        else:
            # string stored in field buffers
            return self._memory.view_field_memory(address, length)

    def check_modify(self, length, address):
        """Assign a new string into an existing buffer."""
        # if it is a code literal, we now do need to allocate space for a copy
        if address >= self._memory.code_start and address < self._memory.var_start():
            length, address = self.store(self.view(length, address).tobytes())
        return length, address

    def store(self, in_str, address=None, check_free=True):
        """Store a new string and return the string pointer."""
        length = len(in_str)
        # don't store overlong strings
        if length > 255:
            raise error.BASICError(error.STRING_TOO_LONG)
        # don't store if address is provided (code or FIELD strings)
        if address is None:
            # reserve string space; collect garbage if necessary
            if check_free:
                self._memory.check_free(length, error.OUT_OF_STRING_SPACE)
            # find new string address
            self.current -= length
            address = self.current + 1
            # don't store empty strings
            if length > 0:
                # copy and convert to bytearray
                self._strings[address] = bytearray(in_str)
        return length, address

    def _delete_last(self):
        """Delete the string provided if it is at the top of string space."""
        last_address = self.current + 1
        try:
            length = len(self._strings[last_address])
            self.current += length
            del self._strings[last_address]
        except KeyError: # pragma: no cover
            # maybe happens if we're called before an out-of-memory exception is handled
            # and the string wasn't allocated
            pass

    def collect_garbage(self, string_ptrs):
        """Re-store the strings referenced in string_ptrs, delete the rest."""
        # string_ptrs should be a list of memoryviews to the original pointers
        # retrieve addresses and copy strings
        string_list = []
        # find last non-temporary string
        last_permanent = self._memory.stack_start()
        last_perm_view = None
        for view in string_ptrs:
            length, addr = struct.unpack('<BH', view.tobytes())
            # exclude empty elements of string arrays (len==0 and addr==0)
            # exclude strings is not located in memory (FIELD or code strings)
            if addr >= self._memory.var_start():
                string_list.append((view, addr, self._retrieve(length, addr)))
                # set sentinel string (lowest-address permanent string)
                # don't use zero-length strings as sentinel:
                # they share an address with allocated strings and may get swapped on sorting
                # in which case the allocated permanent string ends up below the sentinel
                if self._temp is not None and length > 0:
                    if addr > self._temp and addr < last_permanent:
                        last_permanent, last_perm_view = addr, view
        # sort by address, largest first (maintain order of storage)
        string_list.sort(key=itemgetter(1), reverse=True)
        # clear the string buffer and re-store all referenced strings
        self.clear()
        for view, _, string in string_list:
            # re-allocate string space
            # update the original pointers supplied (these are memoryviews)
            view[:] = struct.pack('<BH', *self.store(string, check_free=False))
        # readdress  start of temporary strings
        if last_perm_view is None:
            self._temp = None
        elif self._temp is not None and self._temp != self._memory.stack_start():
            self._temp = -1 + struct.unpack_from('<H', last_perm_view.tobytes(), 1)[0]

    def get_memory(self, address):
        """Retrieve data from data memory: string space """
        # find the variable we're in
        for try_address, value in iteritems(self._strings):
            length = len(value)
            if try_address <= address < try_address + length:
                return value[address - try_address]
        return -1

    def fix_temporaries(self):
        """Make all temporary strings permanent."""
        self._temp = self.current

    def reset_temporaries(self):
        """Delete temporary string at top of string space."""
        if self._temp is not None and self._temp != self.current:
            self._delete_last()
        self._temp = self.current

    def is_permanent(self, string):
        """Return whether string is in permanent string space."""
        addr = string.address()
        return addr > self._temp

    def is_field_string(self, string):
        """Return whether string is a FIELD string."""
        return string.address() < self._memory.code_start



"""
PC-BASIC - display.screencopyhandler
Clipboard copy & print screen handler

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ..base import scancode
from ..base import signals


class ScreenCopyHandler(object):
    """Event handler for clipboard copy and print screen."""

    # handle an input signal, read the screen
    # and write the text to an output queue or file
    # independently of what BASIC is doing

    def __init__(self, queues, text_screen, lpt1_file):
        """Initialise copy handler."""
        self._queues = queues
        self._text_screen = text_screen
        self._lpt1_file = lpt1_file

    def check_input(self, signal):
        """Handle input signals."""
        if signal.event_type == signals.CLIP_COPY:
            self._copy_clipboard(*signal.params)
            return True
        elif signal.event_type == signals.KEYB_DOWN:
            c, scan, mod = signal.params
            if scan == scancode.PRINT and (scancode.LSHIFT in mod or scancode.RSHIFT in mod):
                # shift+printscreen triggers a print screen
                self._print_screen(self._lpt1_file)
                return True
        return False

    def _print_screen(self, target_file):
        """Output the visible page to file in raw bytes."""
        if not target_file:
            return
        for line in self._text_screen.get_chars(as_type=bytes):
            target_file.write(b''.join(line).replace(b'\0', b' '))
        target_file.write_line()

    def _copy_clipboard(self, start_row, start_col, stop_row, stop_col):
        """Copy selected screen area to clipboard."""
        # in the signal, stop_row is inclusive but stop_col is *exclusive* ?
        text = list(self._text_screen.get_text(
            start_row=start_row, start_col=start_col, stop_row=stop_row, stop_col=stop_col-1
        ))
        clip_text = u'\n'.join(u''.join(_row) for _row in text)
        self._queues.video.put(signals.Event(
            signals.VIDEO_SET_CLIPBOARD_TEXT, (clip_text,)
        ))



"""
PC-BASIC - compat
Cross-platform compatibility utilities

(c) 2018--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

# pylint: disable=no-name-in-module, import-error, undefined-variable, used-before-assignment

import sys
import os
import io


##################################################################################################
# not available in Python <= 3.6

try:
    import importlib_resources as resources
except ImportError: # pragma: no cover
    from importlib import resources

try:
    from contextlib import nullcontext
except ImportError: # pragma: no cover

    class nullcontext():
        def __init__(self, enter_result=None):
            self.enter_result = enter_result
        def __enter__(self):
            return self.enter_result
        def __exit__(self, *excinfo):
            pass

##################################################################################################


from .base import PLATFORM, PY2, WIN32, MACOS, X64
from .base import USER_CONFIG_HOME, USER_DATA_HOME, BASE_DIR, HOME_DIR
from .streams import StreamWrapper, fix_stdio, is_readable_text_stream, is_writable_text_stream

from .console import console, read_all_available, IS_CONSOLE_APP
from .console import stdio, init_stdio


if PY2: # pragma: no cover
    from .python2 import add_str, iterchar
    from .python2 import xrange, zip, iteritems, itervalues, iterkeys, iterbytes
    from .python2 import getcwdu, getenvu, setenvu, iterenvu
    from .python2 import configparser, queue, copyreg, which
    from .python2 import SimpleNamespace, TemporaryDirectory
    from .python2 import BrokenPipeError, is_broken_pipe
    unichr, int2byte, text_type = unichr, chr, unicode

    if WIN32:
        from . import win32_subprocess
        from .win32 import argv
    else:
        from .posix import argv
else:
    import configparser, queue, copyreg
    from shutil import which
    from types import SimpleNamespace
    from tempfile import TemporaryDirectory
    from .python3 import int2byte, add_str, iterchar, iterbytes
    from .python3 import xrange, zip, iteritems, itervalues, iterkeys
    from .python3 import getcwdu, getenvu, setenvu, iterenvu
    from .python3 import is_broken_pipe
    BrokenPipeError = BrokenPipeError
    unichr, text_type = chr, str
    argv = sys.argv


if WIN32:
    from .win32 import set_dpi_aware, line_print
    from .win32 import get_free_bytes, get_short_pathname, is_hidden
    from .win32 import EOL, EOF
    from .win32 import SHELL_ENCODING, OEM_ENCODING, HIDE_WINDOW
else:
    from .posix import set_dpi_aware, line_print
    from .posix import get_free_bytes, get_short_pathname, is_hidden
    from .posix import EOL, EOF
    from .posix import SHELL_ENCODING, OEM_ENCODING, HIDE_WINDOW


if MACOS:
    # on MacOS, if launched from Finder, ignore the additional "process serial number" argument
    argv = [_arg for _arg in argv if not _arg.startswith('-psn_')]
    # for macOS - if no console, presumably we're launched as a bundle
    # set working directory to user home
    # bit of a hack but I don't know a better way
    if not IS_CONSOLE_APP:
        os.chdir(HOME_DIR)


##################################################################################################
# fix stdio and deal with broken pipes in scripts

from contextlib import contextmanager

@contextmanager
def script_entry_point_guard():
    """Wrapper for entry points, to deal with stdio, Ctrl-C and sigpipe."""
    fix_stdio()
    init_stdio()
    # see docs.python.org/3/library/signal.html#note-on-sigpipe
    # for cases where shell tools send SIGPIPE
    # e.g. echo -e "?1\r?2\r" | python3.8 -m pcbasic -n | head --lines=1
    exit_code = True
    try:
        yield
        exit_code = False
    except KeyboardInterrupt:
        exit_code = False
    except BrokenPipeError as e:
        # py2 hack
        if not is_broken_pipe(e):
            raise
    # broken pipe usually gets caught above, but flush streams here as a failsafe
    try:
        sys.stdout.flush()
    except Exception:
        exit_code = True
    try:
        sys.stderr.flush()
    except Exception:
        exit_code = True
    if exit_code:
        try:
            os.dup2(os.open(os.devnull, os.O_WRONLY), sys.stdout.fileno())
        except Exception:
            pass
        try:
            os.dup2(os.open(os.devnull, os.O_WRONLY), sys.stderr.fileno())
        except Exception:
            pass
    sys.exit(exit_code)



##################################################################################################
# utility functions, this has to go somewhere...

import re
import random

def _build_split_regexp(split_by, quote, as_type):
    """
    Build regexp for use with split_quoted and split_pair.
    `split_by` and `quote` must be of type `as_type`.
    """
    if not quote:
        quote = as_type()
    quote = re.escape(quote)
    if split_by is None:
        # by default, split on whitespace
        split_by = u'\s'
    else:
        split_by = re.escape(split_by)
    # https://stackoverflow.com/questions/16710076/python-split-a-string-respect-and-preserve-quotes
    # note ur'' is not accepted by python 3, and r'' means bytes in python2.
    # bytes has no .format so using % which is awkward here
    pattern = (
        br'(?:[^{%s}{%s}]|[{%s}](?:\\.|[^{%s}])*[{%s}])+'
    )
    if as_type == type(u''):
        # we know the template pattern string and ascii is ok
        pattern = pattern.decode('ascii', 'ignore')
    regexp = pattern % (split_by, quote, quote, quote, quote)
    return regexp

def split_quoted(line, split_by=None, quote=None, strip_quotes=False):
    """
    Split by separators, preserving quoted blocks; \\ escapes quotes.
    """
    regexp = _build_split_regexp(split_by, quote, as_type=type(line))
    chunks = re.findall(regexp, line)
    if strip_quotes:
        chunks = [c.strip(quote) for c in chunks]
    return chunks

def split_pair(line, split_by=None, quote=None):
    """
    Split by separators, preserving quoted blocks; \\ escapes quotes.
    First match only, always return two values.
    """
    regexp = _build_split_regexp(split_by, quote, as_type=type(line))
    for match in re.finditer(regexp, line):
        s0 = match.group()
        s1 = line[match.end()+1:]
        # only loop once
        return s0, s1

def iter_chunks(char_list, attrs):
    """Iterate over list yielding chunks of elements with the same attribute."""
    last_attr = None
    chars = []
    # collect chars in chunks with the same attribute
    for char, attr in zip(char_list, attrs):
        if attr != last_attr:
            if last_attr is not None:
                yield chars, last_attr
            last_attr = attr
            chars = []
        chars.append(char)
    if chars:
        yield chars, attr

def random_id(number_digits, prefix='', exclude=()):
    """Generate a random hex id as bytes, optionally exclude from a given set."""
    num_ids = 10**number_digits
    # construct the template for the next % operation, e.g. '07X' if number_digits == 7
    format_spec = '0{}X'.format(number_digits)
    for _ in xrange(num_ids):
        name = format(random.randint(0, num_ids), format_spec)
        if isinstance(prefix, bytes):
            name = name.encode('ascii')
        name = prefix + name
        if name not in exclude:
            return name
    raise RuntimeError('no free id of length {} available'.format(number_digits))



"""
PC-BASIC - arrays.py
Array variable management

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import binascii
import struct

from ...compat import iteritems, iterkeys

from ..base import error
from .. import values
from .scalars import get_name_in_memory


class Arrays(object):

    def __init__(self, memory, values):
        """Initialise arrays."""
        self._memory = memory
        self._values = values
        self.clear()
        self.clear_base()

    def __contains__(self, varname):
        """Check if a scalar has been defined."""
        return varname in self._dims

    def __iter__(self):
        """Return an iterable over all scalar names."""
        return iterkeys(self._dims)

    def __repr__(self):
        """Debugging representation of variable dictionary."""
        return '\n'.join(
            '%s%s: %s' % (
                n.decode('ascii'),
                v,
                binascii.hexlify(bytes(self._buffers[n])).decode('ascii')
            )
            for n, v in iteritems(self._dims)
        )

    def clear(self):
        """Clear arrays."""
        self._dims = {}
        self._buffers = {}
        self._array_memory = {}
        self.current = 0

    def erase_(self, args):
        """Remove an array from memory."""
        for name in args:
            name = self._memory.complete_name(name)
            if name not in self._dims:
                # IFC if array does not exist
                raise error.BASICError(error.IFC)
            dimensions = self._dims[name]
            record_len = 1 + max(3, len(name)) + 3 + 2*len(dimensions)
            freed_bytes = self._buffer_size(name, dimensions) + record_len
            erased_name_ptr, _ = self._array_memory[name]
            # delete buffers
            del self._dims[name]
            del self._buffers[name]
            del self._array_memory[name]
            # update memory model
            for name in self._array_memory:
                name_ptr, array_ptr = self._array_memory[name]
                if name_ptr > erased_name_ptr:
                    self._array_memory[name] = name_ptr - freed_bytes, array_ptr - freed_bytes
            self.current -= freed_bytes
        # if all arrays have been cleared and array base was set to 0 implicitly by DIM, unset it
        # however, if array base was set explicitly by OPTION BASE, it remains set.
        if not self._dims and self._base_set_by_dim:
            self.clear_base()

    def index(self, index, dimensions):
        """Return the flat index for a given dimensioned index."""
        bigindex = 0
        area = 1
        for i in range(len(index)):
            # dimensions is the *maximum index number*, regardless of self._base
            bigindex += area * (index[i] - self._base)
            area *= dimensions[i] + 1 - self._base
        return bigindex

    def view_full_buffer(self, name):
        """Return a memoryview to a full array."""
        return memoryview(self._buffers[name])

    def dimensions(self, name):
        """Return the dimensions of an array."""
        return self._dims[name]

    def dim_(self, args):
        """DIM: dimension arrays."""
        for a in args:
            name, indices = a
            self.allocate(self._memory.complete_name(name), indices)

    @staticmethod
    def _record_size(name, dimensions):
        """Calculate size of array record in bytes."""
        # first two bytes: chars of name or 0 if name is one byte long
        return 1 + max(3, len(name)) + 3 + 2*len(dimensions)

    def flat_length(self, dimensions):
        """Total number of elements."""
        return self.index(dimensions, dimensions) + 1

    def _buffer_size(self, name, dimensions):
        """Calculate size of array buffer in bytes."""
        return self.flat_length(dimensions) * values.size_bytes(name)

    def memory_size(self, name, dimensions):
        """Calculate size of array record and buffer in bytes."""
        return self._record_size(name, dimensions) + self._buffer_size(name, dimensions)

    def allocate(self, name, dimensions):
        """
        Allocate array space for an array of given dimensioned size.
        Raise errors if duplicate name or illegal index value.
        """
        if not dimensions:
            # DIM A does nothing
            return
        if name in self._dims:
            raise error.BASICError(error.DUPLICATE_DEFINITION)
        # a call that raises ifc does not implicitly set the array base
        if any(_d < 0 for _d in dimensions):
            raise error.BASICError(error.IFC)
        # implicitly set array base
        if self._base is None:
            self._base = 0
            self._base_set_by_dim = True
        elif any(_d < self._base for _d in dimensions):
            raise error.BASICError(error.SUBSCRIPT_OUT_OF_RANGE)
        # update memory model
        name_ptr = self.current
        record_len = self._record_size(name, dimensions)
        array_bytes = self._buffer_size(name, dimensions)
        array_ptr = name_ptr + record_len
        total_bytes = record_len + array_bytes
        self._memory.check_free(total_bytes, error.OUT_OF_MEMORY)
        self.current += total_bytes
        self._array_memory[name] = (name_ptr, array_ptr)
        self._buffers[name] = bytearray(array_bytes)
        self._dims[name] = dimensions

    def check_dim(self, name, index):
        """
        Check if an array has been allocated.
        If not, auto-allocate if indices are <= 10; raise error otherwise.
        """
        try:
            dimensions = self._dims[name]
        except KeyError:
            # auto-dimension - 0..10 or 1..10
            # this even fixes the dimensions if the index turns out to be out of range
            dimensions = [10] * len(index)
            self.allocate(name, dimensions)
        lst = self._buffers[name]
        if len(index) != len(dimensions):
            raise error.BASICError(error.SUBSCRIPT_OUT_OF_RANGE)
        for i, d in zip(index, dimensions):
            if i < 0:
                raise error.BASICError(error.IFC)
            elif i < self._base or i > d:
                # dimensions is the *maximum index number*, regardless of self._base
                raise error.BASICError(error.SUBSCRIPT_OUT_OF_RANGE)
        return dimensions, lst

    def clear_base(self):
        """Unset the array base."""
        # OPTION BASE value. NONE: unset
        self._base = None
        # OPTION BASE set by DIM rather than explicitly
        self._base_set_by_dim = False

    def option_base_(self, args):
        """Set the array base to 0 or 1 (OPTION BASE). Raise error if already set."""
        base, = args
        base = int(base)
        if self._base is not None and base != self._base:
            # duplicate definition
            raise error.BASICError(error.DUPLICATE_DEFINITION)
        self._base = base

    def view_buffer(self, name, index):
        """Return a memoryview to an array element."""
        dimensions, lst = self.check_dim(name, index)
        bigindex = self.index(index, dimensions)
        bytesize = values.size_bytes(name)
        return memoryview(lst)[bigindex*bytesize:(bigindex+1)*bytesize]

    def get(self, name, index):
        """Retrieve a view of the value of an array element."""
        # do not make a copy - we may end up with stale string pointers
        # due to garbage collection
        return self._values.create(self.view_buffer(name, index))

    def set(self, name, index, value):
        """Assign a value to an array element."""
        if isinstance(value, values.String):
            self._memory.strings.fix_temporaries()
        # copy value into array
        self.view_buffer(name, index)[:] = values.to_type(name[-1:], value).to_bytes()
        # drop cache here

    def varptr(self, name, indices):
        """Retrieve the address of an array."""
        dimensions = self._dims[name]
        _, array_ptr = self._array_memory[name]
        # arrays are kept at the end of the var list
        return (
            self._memory.var_current() + array_ptr +
            values.size_bytes(name) * self.index(indices, dimensions)
        )

    def dereference(self, address):
        """Get a value for an array given its pointer address."""
        found_addr = -1
        found_name = None
        for name, data in iteritems(self._array_memory):
            addr = self._memory.var_current() + data[1]
            if addr > found_addr and addr <= address:
                found_addr = addr
                found_name = name
        if not found_name:
            return None
        lst = self._buffers[name]
        offset = address - found_addr
        return self._values.from_bytes(lst[offset : offset+values.size_bytes(name)])

    def get_memory(self, address):
        """Retrieve data from data memory: array space """
        name_addr = -1
        arr_addr = -1
        for name in self._array_memory:
            name_try, arr_try = self._array_memory[name]
            if name_try <= address and name_try > name_addr:
                name_addr, arr_addr = name_try, arr_try
                the_arr = name
                break
        else: # pragma: no cover
            return -1
        var_current = self._memory.var_current()
        dimensions = self._dims[the_arr]
        if address >= var_current + arr_addr:
            offset = address - arr_addr - var_current
            if offset >= self._buffer_size(the_arr, dimensions): # pragma: no cover
                return -1
            byte_array = self._buffers[the_arr]
            return ord(byte_array[offset:offset+1])
        else:
            offset = address - name_addr - var_current
            if offset < max(3, len(the_arr))+1:
                return get_name_in_memory(the_arr, offset)
            else:
                offset -= max(3, len(the_arr))+1
                data_rep = struct.pack(
                    '<HB',
                    self._buffer_size(the_arr, dimensions) + 1 + 2*len(dimensions),
                    len(dimensions)
                )
                for d in dimensions:
                    data_rep += struct.pack('<H', d + 1 - self._base)
                return ord(data_rep[offset:offset+1])

    def get_strings(self):
        """Return a list of views of string array elements."""
        return [
            memoryview(buf)[i:i+3]
            for name, buf in iteritems(self._buffers)
            if name[-1:] == values.STR
            for i in range(0, len(buf), 3)
        ]


    ###########################################################################
    # helper functions for Python interface

    def from_list(self, python_list, name):
        """Convert Python list to BASIC array."""
        self._from_list(python_list, name, [])

    def _from_list(self, python_list, name, index):
        """Convert Python list to BASIC array."""
        if not python_list:
            raise ValueError('Array must not be empty.')
        if isinstance(python_list[0], list):
            for i, v in enumerate(python_list):
                self._from_list(v, name, index+[i+(self._base or 0)])
        else:
            for i, v in enumerate(python_list):
                self.set(name, index+[i+(self._base or 0)], self._values.from_value(v, name[-1:]))

    def to_list(self, name):
        """Convert BASIC array to Python list."""
        if name not in self._dims:
            return []
        indices = self._dims[name]
        return self._to_list(name, [], indices)

    def _to_list(self, name, index, remaining_dimensions):
        """Convert BASIC array to Python list."""
        if len(remaining_dimensions) == 1:
            return [
                self.get(name, index+[i]).to_value()
                for i in range((self._base or 0), remaining_dimensions[0] + 1)
            ]
        else:
            return [
                self._to_list(name, index+[i], remaining_dimensions[1:])
                for i in range((self._base or 0), remaining_dimensions[0] + 1)
            ]



"""
PC-BASIC - data.programs
Bundled BASIC programs

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ...compat import resources


# list of available programs
PROGRAMS = tuple(
    name
    for name in resources.contents(__package__)
    if name.lower().endswith('.bas') and resources.is_resource(__package__, name)
)


def read_program_file(name):
    """Read a bundled BASIC program file."""
    return resources.read_binary(__package__, name)
